/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "AudioDeviceInterface.h"
#include "AudioDeviceInterface_p.h"

#include "AudioDeviceSocket.h"
#include "AudioDeviceSpeed.h"
#include "CppUtil.h"
#include "CddbQuery.h"

#include "XmlOptions.h"

#include <QDir>
#include <QHostAddress>
#include <QTcpServer>
#include <QTcpSocket>
#include <QUrl>


// Copied from asunder v2.2 - GPL v2
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>

#if defined(__FreeBSD__) || defined(__FreeBSD_kernel__)
#include <sys/cdio.h>
#include <arpa/inet.h>
#elif defined(__linux__)
#include <linux/cdrom.h>
#endif

#include <array>

//_________________________________________
QUrl AudioDeviceInterface::Private::url( const TrackInformation& trackInformation ) const
{

    // get url from track
    const auto extraItems = Base::makeT<Base::QueryItemList>( {{"device", device_}} );
    QUrl url( trackInformation.url( extraItems ) );

    // set scheme, host, port and path
    url.setScheme("http");
    url.setHost( QHostAddress( QHostAddress::LocalHost ).toString() );
    url.setPort( port_ );
    url.setPath( QString( "/musicbox-audiostream/%1/%2.wav" ).arg( device_ ).arg( trackInformation.track() ) );

    return url;

}

//______________________________________________________________
void AudioDeviceInterface::Private::createTrackList()
{

    Debug::Throw( "AudioDeviceInterface::Private::createTrackList.\n" );

    // clear existing tracks
    trackList_.clear();

    #if WITH_CDIO_PARANOIA
    for( int iTrack = 0; iTrack < cddbDiscInfo_.frameOffsets_.size()-1; ++iTrack )
    {

        const int trackId = iTrack + cddbDiscInfo_.firstTrack_;

        // create new track
        TrackInformation track;
        track.setTitle( tr( "Track %1" ).arg( trackId ) );
        track.setTrack( trackId );
        track.setTrackId( iTrack );
        track.setTrackIndex( trackId );
        track.setDuration( ( cddbDiscInfo_.frameOffsets_[iTrack+1] - cddbDiscInfo_.frameOffsets_[iTrack] )/CDIO_CD_FRAMES_PER_SEC );
        track.setFile( File( url( track ).toEncoded() ) );

        // append to list
        trackList_.append( track );

    }
    #endif

}

//______________________________________________________________
void AudioDeviceInterface::Private::createCddbDisc()
{

    Debug::Throw( "AudioDeviceInterface::Private::createCddbDisc.\n" );

    #if WITH_CDDB && WITH_CDIO_PARANOIA

    // cddb disc deleter
    class CddbDiscDeleter
    {
        public:
        void operator()( cddb_disc_t* disc ) const
        { cddb_disc_destroy( disc ); }
    };

    // create cd and check
    cddbDisc_ = CddbQuery::CddbDiscPointer( cddb_disc_new(), CddbDiscDeleter() );
    auto cddbDisc( cddbDisc_.get() );
    if( !cddbDisc ) return;

    for( int iTrack = 0; iTrack < cddbDiscInfo_.frameOffsets_.size()-1; ++iTrack )
    {

        const int trackId = iTrack + cddbDiscInfo_.firstTrack_;

        // create cddb track
        auto track = cddb_track_new();
        if( !track ) continue;

        cddb_track_set_frame_offset(track, cddbDiscInfo_.frameOffsets_[iTrack] );
        cddb_track_set_title(track, tr( "Track %1" ).arg( trackId ).toUtf8().constData() );
        cddb_track_set_artist(track, TrackInformation::UnknownArtist.toUtf8().constData() );
        cddb_disc_add_track(cddbDisc, track);

    }

    // disc attributes
    cddb_disc_set_artist(cddbDisc, TrackInformation::UnknownArtist.toUtf8().constData() );
    cddb_disc_set_title(cddbDisc, TrackInformation::UnknownAlbum.toUtf8().constData() );
    cddb_disc_set_length( cddbDisc, cddbDiscInfo_.frameOffsets_.back() / CDIO_CD_FRAMES_PER_SEC );
    cddb_disc_calc_discid( cddbDisc );

    // debug output
    if( Debug::level() > 0 ) cddb_disc_print(cddbDisc);

    #endif

}

//______________________________________________________________
AudioDeviceInterface::AudioDeviceInterface( QObject* parent ):
    QObject( parent ),
    Counter( "AudioDeviceInterface" ),
    d( new Private )
{
    Debug::Throw( "AudioDeviceInterface::AudioDeviceInterface" );

    #if WITH_CDDB
    // make sure cover path is saved
    QDir().mkpath( XmlOptions::get().raw( "CDDB_CACHE_DIR" ) );
    #endif

}

//______________________________________________________________
AudioDeviceInterface::~AudioDeviceInterface() = default;

//______________________________________________________________
void AudioDeviceInterface::initializeServer()
{
    Debug::Throw( "AudioDeviceInterface::initializeServer.\n" );

    d->server_.reset( new QTcpServer( this ) );
    connect( d->server_.get(), SIGNAL(newConnection()), SLOT(_newConnection()) );

    // try get port from options
    if( XmlOptions::get().contains( "AUDIO_DEVICE_SERVER_PORT" ) ) d->port_ = XmlOptions::get().get<int>( "AUDIO_DEVICE_SERVER_PORT" );
    else d->port_ = 0;

    if( !( d->port_ > 0 && d->server_->listen( QHostAddress::LocalHost, d->port_ ) ) )
    {
        d->server_->listen( QHostAddress::LocalHost, 0 );
        d->port_ = d->server_->serverPort();
        XmlOptions::get().set<int>( "AUDIO_DEVICE_SERVER_PORT", d->port_ );
    }

}

//______________________________________________________________
void AudioDeviceInterface::tryReadDevice(const QString& device)
{
    if( !device.isEmpty() && d->device_ != device )
    { readDevice( device ); }
}

//______________________________________________________________
void AudioDeviceInterface::readDevice(const QString& device)
{

    Debug::Throw( "AudioDeviceInterface::readDevice.\n" );

    // initialize server
    if( !d->server_ ) initializeServer();

    // reset
    _reset();

    // assign device
    d->device_ = device;

    // try read information from device
    if( _readDisc() )
    {

        emit tracksFound( d->trackList_ );

        // cddb query
        #if WITH_CDDB && WITH_CDIO_PARANOIA
        auto query = new CddbQuery( this );
        connect( query, SIGNAL(tracksFound(TrackInformation::List)), SLOT(_updateTracks(TrackInformation::List)) );
        connect( query, SIGNAL(finished()), query, SLOT(deleteLater()) );

        query->setCddbDisc( d->cddbDisc_ );
        query->start();
        #endif

    } else {

        emit invalidDevice( device );
        _reset();

    }

}

//______________________________________________________________
bool AudioDeviceInterface::tryInitCdio()
{
    #if WITH_CDIO_PARANOIA
    if( !d->device_.isEmpty() && d->drive_ && d->paranoia_ ) return true;
    if( !d->device_.isEmpty() ) return _initCdio();
    else return false;
    #else
    return false;
    #endif
}

//______________________________________________________________
void AudioDeviceInterface::saveElapsedTime( quint32 value )
{
    Debug::Throw() << "AudioDeviceInterface::saveElapsedTime - elapsed time: " << value << endl;
    d->elapsedTime_ = value;
}

//______________________________________________________________
void AudioDeviceInterface::_newConnection()
{
    Debug::Throw( "AudioDeviceInterface::_newConnection.\n" );
    while( d->server_->hasPendingConnections() )
    { new AudioDeviceSocket( this, d->server_->nextPendingConnection() ); }
}

//______________________________________________________________
void AudioDeviceInterface::_updateTracks( const TrackInformation::List& trackList )
{

    for( int iTrack = 0; iTrack < d->trackList_.size() && iTrack < trackList.size(); ++iTrack )
    {
        // get track
        auto track( trackList[iTrack] );
        track.setFile( File( d->url( track ).toEncoded() ) );

        // emit signal
        emit trackUpdated( d->trackList_[iTrack], track );

        // replace in old list
        d->trackList_[iTrack] = track;
    }

}

//______________________________________________________________
bool AudioDeviceInterface::_readDisc()
{

    Debug::Throw( "AudioDeviceInterface::_readDisc.\n" );
    const auto device( d->device_ );

    FileDescriptor fd( open( device.toLatin1(), O_RDONLY | O_NONBLOCK ) );
    if( fd < 0 ) return false;

    #if defined(__linux__) && WITH_CDIO_PARANOIA

    // check disc status
    auto status = ioctl( fd, CDROM_DISC_STATUS, CDSL_CURRENT );
    if( !( status == CDS_AUDIO || status == CDS_MIXED ) ) return false;

    // read toc header
    struct cdrom_tochdr header;
    if( ioctl(fd, CDROMREADTOCHDR, &header) != 0 ) return false;

    // fill cddbDisc info
    CddbDiscInfo cddbDiscInfo;
    cddbDiscInfo.firstTrack_ = header.cdth_trk0;
    cddbDiscInfo.lastTrack_ = header.cdth_trk1;

    // cantata adds a static 2s offsets to all frame offsets and to disc length
    static const auto baseFrameOffset = 2*CDIO_CD_FRAMES_PER_SEC;

    // cdrom entry structure
    struct cdrom_tocentry entry;
    entry.cdte_format = CDROM_LBA;

    // loop over tracks
    for( int iTrack = header.cdth_trk0; iTrack <= header.cdth_trk1; ++iTrack )
    {

        // read toc entry
        entry.cdte_track = iTrack;
        if( ioctl( fd, CDROMREADTOCENTRY, &entry ) == 0 )
        {

            cddbDiscInfo.frameOffsets_.append( baseFrameOffset + entry.cdte_addr.lba );

        } else return false;

    }

    // get disc length
    entry.cdte_track = CDROM_LEADOUT;
    if( ioctl(fd, CDROMREADTOCENTRY, &entry ) == 0 )
    {

        cddbDiscInfo.frameOffsets_.append( baseFrameOffset + entry.cdte_addr.lba );

    } else return false;

    // assign disc info
    d->cddbDiscInfo_ = cddbDiscInfo;
    d->createTrackList();
    d->createCddbDisc();
    return true;

    #else

    return false;

    #endif


}

//______________________________________________________________
bool AudioDeviceInterface::_initCdio()
{

    Debug::Throw( "AudioDeviceInterface::_initCdio.\n" );

    #if WITH_CDIO_PARANOIA
    const auto device( d->device_ );
    d->drive_.reset( cdio_cddap_identify( device.toLatin1().data(), 0, nullptr ) );
    if( !d->drive_ ) return false;

    // try open drive
    cdio_cddap_open(d->drive_.get());

    // limit speed
    auto speedMode = static_cast<AudioDeviceSpeed::SpeedMode>( XmlOptions::get().get<int>( "AUDIO_DEVICE_SPEED" ) );
    if( speedMode != AudioDeviceSpeed::SpeedMode::Default )
    { cdio_cddap_speed_set(d->drive_.get(), AudioDeviceSpeed::speed( speedMode ) ); }

    // create paranoia structure and check
    d->paranoia_.reset( cdio_paranoia_init( d->drive_.get() ) );

    // set paranoia mode
    cdio_paranoia_modeset( d->paranoia_.get(), PARANOIA_MODE_FULL^PARANOIA_MODE_NEVERSKIP );

    return static_cast<bool>(d->paranoia_);
    #else
    return false;
    #endif
}


//______________________________________________________________
void AudioDeviceInterface::_reset()
{

    d->device_.clear();
    d->trackList_.clear();
    d->cddbDiscInfo_.clear();

    #if WITH_CDIO_PARANOIA
    d->drive_.reset();
    d->paranoia_.reset();
    #endif

    #if WITH_CDDB
    d->cddbDisc_.reset();
    #endif

}
