/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "AudioDeviceSocket.h"

#include "AudioDeviceInterface.h"
#include "AudioDeviceInterface_p.h"

#include "CppUtil.h"

#if QT_VERSION >= 0x050000
#include <QUrlQuery>
#endif

//______________________________________________________________
namespace Local
{
    static const int wavHeaderSize=44;
}

//______________________________________________________________
AudioDeviceSocket::AudioDeviceSocket( AudioDeviceInterface* parent, QTcpSocket* socket ):
    QObject( parent ),
    Counter( "AudioDeviceSocket" ),
    interface_( parent ),
    socket_( socket )
{
    connect( socket_.get(), SIGNAL(readyRead()), this, SLOT(_read()) );
    connect( socket_.get(), SIGNAL(disconnected()), this, SLOT(deleteLater()) );
}

//______________________________________________________________
void AudioDeviceSocket::_read()
{

    Debug::Throw( "AudioDeviceSocket::_read.\n" );

    class SocketClose: private Base::NonCopyable<SocketClose>
    {
        public:

        //* constructor
        SocketClose( QTcpSocket* socket ):
            socket_( socket )
            {}

        //* destructor
        ~SocketClose()
        { socket_->close(); }

        private:

        QTcpSocket* socket_ = nullptr;

    };

    SocketClose socketClose( socket_.get() );

    // read and parse
    auto data = QString::fromUtf8( socket_->readAll() );
    QRegExp regExp( "GET\\s+(.+)(?:\\s+|\r|\n)" );
    regExp.setMinimal(true);
    if( data.indexOf( regExp ) < 0 ) return;

    // read reply and parse url
    const auto request = regExp.cap(1);
    const auto url = QUrl::fromEncoded( request.toUtf8() );

    // parse track information from url
    const TrackInformation trackInformation( url );

    // parse device from url
    #if QT_VERSION >= 0x050000
    const QUrlQuery query( url );
    #else
    const QUrl& query( url );
    #endif

    if( !query.hasQueryItem( "device" ) )
    {
        Debug::Throw(0) << "AudioDeviceSocket::_read - no device" << endl;
        return;
    }
    const auto device =  query.queryItemValue( "device" );

    //* check device
    if( device.isEmpty() )
    {
        Debug::Throw(0) << "AudioDeviceSocket::_read - invalid device" << endl;
        return;
    }

    //* try read device
    interface_->tryReadDevice( device );

    //* check track duration and offset against stored
    const auto& trackList( interface_->d->trackList_ );
    const auto iter = std::find_if( trackList.begin(), trackList.end(), TrackInformation::SameTrackFTor( trackInformation ) );
    if( iter == trackList.end() || iter->duration() != trackInformation.duration() )
    {
        Debug::Throw(0) << "AudioDeviceSocket::_read - invalid track" << endl;
        return;
    }

    #if WITH_CDIO_PARANOIA

    // try initialize cdio for this device
    if( !interface_->tryInitCdio() )
    {
        Debug::Throw(0) << "AudioDeviceSocket::_read - failed to init cdio" << endl;
        return;
    }

    auto drive = interface_->d->drive_.get();
    auto paranoia = interface_->d->paranoia_.get();

    const auto track( trackInformation.track() );

    // const auto firstSector( cdio_cddap_track_firstsector( drive, track ) );
    const qint32 firstSector( cdio_cddap_track_firstsector( drive, track ) + interface_->d->elapsedTime_ * CDIO_CD_FRAMES_PER_SEC );
    const qint32 lastSector( cdio_cddap_track_lastsector( drive, track ) );
    if( lastSector <= firstSector ) return;

    const qint32 totalSize = ( lastSector - firstSector + 1 )*CD_FRAMESIZE_RAW;

    // write mimetype
    _writeMimeType( QLatin1String( "audio/x-wav" ), totalSize + Local::wavHeaderSize );
    _writeWavHeader( totalSize );

    // seek
    cdio_paranoia_seek( paranoia, firstSector, SEEK_SET);

    // number of written sectors
    int writtenSectors = 0;
    while( firstSector + writtenSectors <= lastSector && socket_->state() == QTcpSocket::ConnectedState )
    {

        auto data = reinterpret_cast<char*>( cdio_paranoia_read_limited( paranoia, nullptr, 1) );
        for( int written = 0; written < CD_FRAMESIZE_RAW; )
        {
            auto tmp = socket_->write( data + written, CD_FRAMESIZE_RAW - written );
            if( tmp < 0 ) break;
            socket_->flush();
            written += tmp;
        }

        // increase sector count
        ++writtenSectors;
    }

    #endif

}


//_________________________________________
bool AudioDeviceSocket::_checkDeviceInterface() const
{
    // check cdio status
    #if WITH_CDIO_PARANOIA
    return interface_ && interface_->d->paranoia_ && interface_->d->drive_ ;
    #else
    return false;
    #endif
}

//_________________________________________________________________
void AudioDeviceSocket::_writeMimeType( const QString &mimeType, qint32 size )
{
    if( mimeType.isEmpty() ) return;
    if( !socket_ ) return;

    QTextStream out( socket_.get() );
    out.setAutoDetectUnicode(true);
    out << "HTTP/1.0 200 OK"
        << "\r\nAccept-Ranges: bytes"
        << "\r\nContent-Length: " << QString::number(size)
        << "\r\nContent-Type: " << mimeType << "\r\n\r\n";

}

//_________________________________________________________________
void AudioDeviceSocket::_writeWavHeader(qint32 size)
{
    if( !socket_ ) return;
    std::array<unsigned char, 44> header =
    {{
        0x52, 0x49, 0x46, 0x46, // 0  "RIFF"
        0x00, 0x00, 0x00, 0x00, // 4  wavSize
        0x57, 0x41, 0x56, 0x45, // 8  "WAVE"
        0x66, 0x6d, 0x74, 0x20, // 12 "fmt "
        0x10, 0x00, 0x00, 0x00, // 16 Size of WAVE section chunk  (ffmpeg has 12 here???)
        0x01, 0x00, 0x02, 0x00, // 20 WAVE type format / Number of channels
        0x44, 0xac, 0x00, 0x00, // 24 Samples per second
        0x10, 0xb1, 0x02, 0x00, // 28 Bytes per second
        0x04, 0x00, 0x10, 0x00, // 32 Block alignment / Bits per sample
        0x64, 0x61, 0x74, 0x61, // 36 "data"  (ffmpeg preceeds this with two 0 bytes)
        0x00, 0x00, 0x00, 0x00  // 40 byteCount
    }};

    auto insertSize = []( unsigned char *data, qint32 size)
    {
        data[0]=(size&0x000000ff);
        data[1]=(size&0x0000ff00)>>8;
        data[2]=(size&0x00ff0000)>>16;
        data[3]=(size&0xff000000)>>24;
    };

    if( size > 0 )
    {

        insertSize( &header[4], size + Local::wavHeaderSize - 8);
        insertSize( &header[40], size );

    }

    socket_->write( reinterpret_cast<const char*>(&header[0]), Local::wavHeaderSize );
}

