/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "AudioDeviceWidget.h"

#include "BrowsedLineEditor.h"
#include "ContextMenu.h"
#include "CppUtil.h"
#include "Debug.h"
#include "IconEngine.h"
#include "IconNames.h"
#include "LineEditor.h"
#include "ListNotificationWidget.h"
#include "PlaylistView.h"
#include "Singleton.h"
#include "TreeView.h"
#include "XmlOptions.h"

#include <QVBoxLayout>
#include <QToolButton>

//____________________________________________________
AudioDeviceWidget::AudioDeviceWidget( QWidget* parent ):
    QWidget( parent ),
    Counter( "AudioDeviceWidget" )
{

    Debug::Throw( "AudioDeviceWidget::AudioDeviceWidget.\n" );

    _installActions();

    auto vLayout = new QVBoxLayout;
    vLayout->setMargin(0);
    vLayout->setSpacing(2);
    setLayout( vLayout );

    deviceEditor_ = new BrowsedLineEditor( this );
    deviceEditor_->setFileMode( QFileDialog::ExistingFile );
    deviceEditor_->setAcceptMode( QFileDialog::AcceptOpen );
    deviceEditor_->editor().setPlaceholderText( tr( "Audio device" ) );
    vLayout->addWidget( deviceEditor_ );

    // update button
    auto updateButton = new QToolButton( deviceEditor_ );
    updateButton->setIcon( IconEngine::get( IconNames::Reload ) );
    updateButton->setAutoRaise( true );
    deviceEditor_->layout()->addWidget( updateButton );

    vLayout->addWidget( list_ = new PlaylistView( this ) );
    list_->setModel( &model_ );
    list_->setOptionName( "AUDIO_DEVICE_LIST" );

    // notification widget
    notificationWidget_ = new ListNotificationWidget( tr( "No media found" ), list_ );
    notificationWidget_->show();

    {
        // context menu
        auto menu = new ContextMenu( list_ );
        menu->addAction( addToPlaylistAction_ );
        menu->addAction( replacePlaylistAction_ );
        connect( menu, SIGNAL(aboutToShow()), SLOT(_updateActions()) );
        menu->setHideDisabledActions( true );
    }

    connect( updateButton, SIGNAL(clicked()), SLOT(_saveDevice()) );
    connect( list_,SIGNAL(activated(QModelIndex)), SLOT(_addItemToPlaylist(QModelIndex)) );
    connect( list_->selectionModel(), SIGNAL(selectionChanged(QItemSelection,QItemSelection)), SLOT(_updateActions()) );

    connect( Base::Singleton::get().application(), SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    _updateConfiguration();

}

//____________________________________________________
void AudioDeviceWidget::tryReadDevice()
{
    Debug::Throw( "AudioDeviceWidget::tryReadDevice.\n" );
    auto device( deviceEditor_->editor().text() );
    if( !device.isEmpty() )
    { emit requestTryReadDevice(device); }
}

//____________________________________________________
void AudioDeviceWidget::showNotificationWidget()
{
    Debug::Throw( "AudioDeviceWidget::showNotificationWidget.\n" );
    notificationWidget_->show();
}

//____________________________________________________
void AudioDeviceWidget::addTracks( const TrackInformation::List& trackList )
{
    Debug::Throw( "AudioDeviceWidget::addTracks.\n" );
    notificationWidget_->setVisible( trackList.isEmpty() );
    model_.set( trackList );
}

//____________________________________________________
void AudioDeviceWidget::updateTrack( const TrackInformation& oldTrack, const TrackInformation& newTrack )
{
    Debug::Throw( "AudioDeviceWidget::updateTrack.\n" );
    auto index( model_.index( oldTrack ) );
    if( index.isValid() ) model_.replace( index, newTrack );
}

//____________________________________________________
void AudioDeviceWidget::_updateConfiguration()
{
    Debug::Throw( "AudioDeviceWidget::_updateConfiguration.\n" );
    deviceEditor_->setFile( XmlOptions::get().raw( "AUDIO_DEVICE" ) );
}

//________________________________________________________
void AudioDeviceWidget::_updateActions()
{
    const auto selectedIndexes( list_->selectionModel()->selectedRows() );
    const bool hasSelection( !selectedIndexes.empty() );
    addToPlaylistAction_->setEnabled( hasSelection );
    replacePlaylistAction_->setEnabled( hasSelection );
}

//____________________________________________________
void AudioDeviceWidget::_saveDevice()
{
    Debug::Throw( "AudioDeviceWidget::_saveDevice.\n" );
    auto device( deviceEditor_->editor().text() );
    XmlOptions::get().setRaw( "AUDIO_DEVICE", device );

    model_.clear();

    if( !device.isEmpty() ) emit requestReadDevice( device );
}

//________________________________________________________
void AudioDeviceWidget::_addItemToPlaylist( const QModelIndex& index )
{
    Debug::Throw( "AudioDeviceWidget::_addItemToPlaylist.\n" );
    if( index.isValid() )
    { emit requestAddToPlaylist( Base::makeT<TrackInformation::List>( { model_.get( index ) } ) ); }
}

//________________________________________________________
void AudioDeviceWidget::_addSelectionToPlaylist()
{
    Debug::Throw( "AudioDeviceWidget::_addSelectionToPlaylist.\n" );
    auto items( model_.get( list_->selectionModel()->selectedRows() ) );
    if( !items.empty() ) emit requestAddToPlaylist( items );
}

//________________________________________________________
void AudioDeviceWidget::_replacePlaylist()
{
    Debug::Throw( "AudioDeviceWidget::_replacePlaylist.\n" );
    auto items( model_.get( list_->selectionModel()->selectedRows() ) );
    if( !items.empty() ) emit requestReplacePlaylist( items );
}

//________________________________________________________
void AudioDeviceWidget::_installActions()
{
    Debug::Throw( "AudioDeviceWidget::_installActions.\n" );

    addAction( addToPlaylistAction_ = new QAction( IconEngine::get( IconNames::Add ), tr( "Add to Playlist" ), this ) );
    connect( addToPlaylistAction_, SIGNAL(triggered()), SLOT(_addSelectionToPlaylist()) );
    addToPlaylistAction_->setEnabled( false );

    addAction( replacePlaylistAction_ = new QAction( tr( "Replace Playlist" ), this ) );
    connect( replacePlaylistAction_, SIGNAL(triggered()), SLOT(_replacePlaylist()) );
    replacePlaylistAction_->setEnabled( false );

}
