/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "CddbQuery.h"

#include "XmlDef.h"
#include "XmlDocument.h"
#include "XmlOptions.h"

#include <QFile>
#include <QUdpSocket>

//_________________________________________
CddbQuery::CddbQuery( QObject* parent ):
    QThread( parent ),
    Counter( "CddbQuery" ),
    host_( XmlOptions::get().get<QString>( "CDDB_HOST" ) ),
    port_( XmlOptions::get().get<int>( "CDDB_PORT" ) )
{}

//_________________________________________
void CddbQuery::run()
{

    #if WITH_CDDB

    // check disc
    if( !cddbDisc_ ) return;

    // try load from disc
    if( _fromDisc() ) return;

    // connection
    if( !_checkConnection() )
    {
        Debug::Throw(0, "CddbQuery::run - connection to cddb server failed (udp).\n" );
        return;
    }

    //* cddb connection deleter
    class CddbConnectionDeleter
    {
        public:
        void operator()( cddb_conn_t* connection ) const
        { cddb_destroy(connection); }
    };

    // cddb connection
    std::unique_ptr<cddb_conn_t, CddbConnectionDeleter> connection( cddb_new() );
    if( !connection ) return;

    // setup connection
    cddb_cache_disable( connection.get() );
    cddb_set_server_name( connection.get(), host_.toUtf8().constData() );
    cddb_set_server_port( connection.get(), port_ );

    // query
    auto cddbDisc( cddbDisc_.get() );
    if( cddb_query( connection.get(), cddbDisc ) <= 0 )
    { return; }

    // read
    do
    {
        if( !cddb_read( connection.get(), cddbDisc ) ) break;

        const auto artist = QString::fromUtf8( cddb_disc_get_artist( cddbDisc ) );
        const auto album = QString::fromUtf8( cddb_disc_get_title( cddbDisc ) );
        const auto year = cddb_disc_get_year( cddbDisc );

        TrackInformation::List trackInformationList;
        const auto trackCount = cddb_disc_get_track_count( cddbDisc );
        for( int iTrack = 0; iTrack < trackCount; ++iTrack )
        {

            // get cddb track
            const auto cddbTrack( cddb_disc_get_track( cddbDisc, iTrack ) );
            if( !cddbTrack ) continue;

            // fill track information
            TrackInformation track;
            track.setArtist( artist );
            track.setAlbum( album );
            track.setYear( year );

            track.setTitle( QString::fromUtf8(cddb_track_get_title( cddbTrack ) ) );
            track.setTrack( cddb_track_get_number( cddbTrack ) );
            track.setDuration( cddb_track_get_length( cddbTrack ) );

            trackInformationList.append( track );
        }

        if( !trackInformationList.empty() )
        {
            _toDisc( trackInformationList );
            emit tracksFound( trackInformationList );
            break;
        }

    } while( cddb_query_next( connection.get(), cddbDisc ) );

    #endif
}

//_________________________________________
bool CddbQuery::_checkConnection() const
{
    Debug::Throw( "CddbQuery::_checkConnection.\n" );
    QUdpSocket socket;
    socket.connectToHost( host_, port_, QIODevice::ReadOnly );
    return socket.waitForConnected(2000);
}

//_________________________________________
bool CddbQuery::_fromDisc()
{
    Debug::Throw( "CddbQuery::_fromDisc.\n" );
    QFile qfile( _fileName() );
    if( !qfile.open( QIODevice::ReadOnly ) )
    { return false; }

    XmlDocument document;
    document.setContent( qfile.readAll() );

    // look for relevant element
    auto topNodes = document.elementsByTagName( Xml::TrackInformationList );
    if( topNodes.isEmpty() ) return false;

    // load track informations
    const auto trackInformationList = TrackInformation::Helper::list( topNodes.at(0).toElement() );
    if( trackInformationList.isEmpty() ) return false;

    // emit
    emit tracksFound( trackInformationList );
    return true;

}

//_________________________________________
void CddbQuery::_toDisc( const TrackInformation::List& trackInformationList ) const
{
    Debug::Throw( "CddbQuery::_toDisc.\n" );
    #if WITH_CDDB

    if( !cddbDisc_ ) return;

    const auto filename( _fileName() );

    // read existing filr if any
    XmlDocument document;
    {
        QFile qfile( filename );
        if( qfile.open( QIODevice::ReadOnly ) )
        { document.setContent( qfile.readAll() ); }
    }

    // get xml node add to document and write
    auto top( TrackInformation::Helper::domElement( trackInformationList, document.get() ) );
    document.replaceChild( top );
    {
        QFile qfile( filename );
        if( qfile.open( QIODevice::WriteOnly ) )  qfile.write( document.toByteArray() );
    }
    #endif
}

//_________________________________________
File CddbQuery::_fileName() const
{
    #if WITH_CDDB
    if( cddbDisc_ ) return File( "Cddb_{" + QString::number( cddb_disc_get_discid( cddbDisc_.get() ) ) + "}.xml" ).addPath( File( XmlOptions::get().raw( "CDDB_CACHE_DIR" ) ) );
    #endif
    return File();
}
