/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "CollectionDelegate.h"

#include "Color.h"
#include "CollectionModel.h"
#include "Singleton.h"
#include "XmlOptions.h"

#include <QApplication>
#include <QStyleOptionViewItem>
#include <QPainter>

#include <algorithm>
#include <numeric>

//_________________________________________________________
CollectionDelegate::CollectionDelegate( QObject *parent ):
    QAbstractItemDelegate( parent ),
    Counter( "CollectionDelegate" )
{

    Debug::Throw( "CollectionDelegate::CollectionDelegate.\n" );

    // configuration
    connect( Base::Singleton::get().application(), SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    _updateConfiguration();

}

//_________________________________________________________
void CollectionDelegate::paint( QPainter *painter, const QStyleOptionViewItem &option, const QModelIndex &index ) const
{

    // check index
    if( !index.isValid() ) return;
    if( index.column() != 0 ) return;

    painter->save();

    // selection
    QStyleOptionViewItemV4 viewOption(option);
    viewOption.showDecorationSelected = true;
    auto style = viewOption.widget ? viewOption.widget->style() : QApplication::style();
    style->drawPrimitive(QStyle::PE_PanelItemViewItem, &viewOption, painter, viewOption.widget);

    // modify color
    QPalette::ColorGroup colorGroup = option.state & QStyle::State_Enabled ? QPalette::Normal : QPalette::Disabled;
    if( colorGroup == QPalette::Normal && !(option.state & QStyle::State_Active) ) colorGroup = QPalette::Inactive;

    QColor foreground = ( option.state & QStyle::State_Selected ) ?
        option.palette.color( colorGroup, QPalette::HighlightedText ):
        option.palette.color( colorGroup, QPalette::Text );

    painter->setPen( foreground );

    // layout icon
    QRect iconRect(
        option.rect.topLeft(),
        QSize( iconSize_, iconSize_ ) );
    iconRect.translate( QPoint( margin_, (option.rect.height() - iconRect.height())/2 ) );
    iconRect = style->visualRect( option.direction, option.rect, iconRect );

    const auto& model( static_cast<const CollectionModel*>( index.model() ) );
    const auto icon = model->data( index, Qt::DecorationRole ).value<QIcon>();

    const auto& item( model->get( index ) );
    const bool needsShadow( !model->icon( item ).isNull() );
    if( needsShadow )
    {
        // drop shadow
        painter->save();
        painter->setPen( QPen( Base::Color( option.palette.color( QPalette::Shadow ) ).addAlpha( 0.3 ).get(), 2 ) );
        painter->setRenderHints(QPainter::Antialiasing );

        auto shadowRect = QRectF( iconRect ).adjusted( 0.5, 0.5, -0.5, -0.5 ).translated( 0.5, 0.5 );
        painter->drawRect( shadowRect );
        painter->restore();
    }

    // icon
    painter->drawPixmap( iconRect, icon.pixmap( QSize( iconSize_, iconSize_ ) ) );

    // text
    const auto text = model->data(index, Qt::DisplayRole ).toString();
    auto textRect( option.rect.adjusted( margin_, margin_, -margin_, -margin_ ) );
    textRect.setLeft( textRect.left() + iconSize_ + spacing_ );
    textRect = style->visualRect( option.direction, option.rect, textRect );
    painter->setFont( option.font );
    painter->drawText( textRect, Qt::AlignLeft|Qt::AlignVCenter, QFontMetrics( option.font ).elidedText( text, Qt::ElideRight, textRect.width() ) );

    painter->restore();

    return;
}

//_________________________________________________________
QSize CollectionDelegate::sizeHint( const QStyleOptionViewItem &option, const QModelIndex &index ) const
{

    if( !index.isValid() ) return QSize();
    if( index.column() != 0 ) return QSize();

    // get model and relevant text
    const auto model( index.model() );
    const auto text = model->data(index, Qt::DisplayRole ).toString();
    auto size = QFontMetrics( option.font ).boundingRect(  text ).size();
    size.setWidth( size.width() + spacing_ + iconSize_ + 2*margin_ );
    size.setHeight( qMax( size.height(), iconSize_ ) + 2*margin_ );
    return size;
}

//________________________________________________________
void CollectionDelegate::_updateConfiguration()
{
    Debug::Throw( "CollectionDelegate::_updateConfiguration.\n" );
    iconSize_ = XmlOptions::get().get<int>( "LIST_ICON_SIZE" );
}
