#ifndef CollectionItem_h
#define CollectionItem_h

/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "Counter.h"
#include "Functors.h"
#include "QOrderedSet.h"
#include "TrackInformation.h"

#include <QList>

class CollectionItem: public TrackInformation
{

    public:

    using List = QList<CollectionItem>;
    using Set = QOrderedSet<CollectionItem>;

    //* used when LogEntry keyword is not defined
    static const QString MimeType;

    //* default constructor
    explicit CollectionItem()
    {}

    //* item type
    enum class Type
    {
        Invalid,
        Artist,
        Album,
        Track
    };

    //* filled constructor
    explicit CollectionItem( Type type, const QString& artist = QString(), const QString& album = QString() ):
        type_( type )
    {
        setArtist( artist );
        setAlbum( album );
    }

    //* constructor from TrackInformation
    explicit CollectionItem( const TrackInformation& other ):
        TrackInformation( other ),
        type_( Type::Track )
    {}

    //* constructor from TrackInformation
    explicit CollectionItem( TrackInformation&& other ):
        TrackInformation( std::move(other) ),
        type_( Type::Track )
    {}

    //* constructor from command reply
    explicit CollectionItem( const QString& value ):
        TrackInformation( value ),
        type_( Type::Track )
    {}

    //*@name accessors
    //@{

    //* item type
    Type type() const
    { return type_; }

    //* parent type
    Type parentType() const
    { return parentType( type() ); }

    //* parent type
    static Type parentType( Type );

    //* true if is children of other item
    bool isChild( const CollectionItem& ) const;

    //* album count
    qint32 albumCount() const
    { return type_ == Type::Artist ? track():0; }

    //* track count
    qint32 trackCount() const
    { return type_ == Type::Album ? track():0; }

    //* true if disc number must be appended to title
    bool showDiscNumber() const
    { return showDiscNumber_; }

    //* hash
    uint hashKey() const override
    {
        switch( type_ )
        {
            case Type::Artist: return qHash( artist().toLower() );
            case Type::Album:  return qHash( (artist()+album() ).toLower() );
            default: return TrackInformation::hashKey();
        }

    }

    //@}

    //*@name modifiers
    //@{

    //* item type
    void setType( Type value )
    { type_ = value; }

    //* album count
    void setAlbumCount( qint32 value )
    { if( type_ == Type::Artist ) setTrack( value ); }

    //* track count
    void setTrackCount( qint32 value )
    { if( type_ == Type::Album ) setTrack( value ); }

    //* show disc number
    void setShowDiscNumber( bool value )
    { showDiscNumber_ = value; }

    //@}

    using SameTypeFTor = Base::Functor::UnaryEqual<CollectionItem, CollectionItem::Type, &CollectionItem::type>;
    using DifferentTypeFTor = Base::Functor::UnaryDiffer<CollectionItem, CollectionItem::Type, &CollectionItem::type>;

    private:

    //* type
    Type type_ = Type::Invalid;

    //* true if disc number must be appended to title
    bool showDiscNumber_ = false;

    //* equal to operator
    friend bool operator == (const CollectionItem& lhs, const CollectionItem& rhs );
    friend bool operator < (const CollectionItem& lhs, const CollectionItem& rhs );

};

#endif
