/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "CollectionModel.h"

#include "CppUtil.h"
#include "IconEngine.h"
#include "IconNames.h"
#include "Singleton.h"
#include "XmlOptions.h"

#include <QMimeData>

//_______________________________________________________________________________________
CollectionModel::CollectionModel( QObject* parent ):
    TreeModel( parent ),
    Counter( "CollectionModel" )
{
    Debug::Throw( "CollectionModel::CollectionModel.\n" );

    // configuration
    connect( Base::Singleton::get().application(), SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    _updateConfiguration();
}

//_______________________________________________________________________________________
Qt::ItemFlags CollectionModel::flags( const QModelIndex& index ) const
{
    return index.isValid() ?
        Qt::ItemIsEnabled | Qt::ItemIsSelectable | Qt::ItemIsDragEnabled:
        static_cast<Qt::ItemFlags>(0);
}

//______________________________________________________________________
QMimeData* CollectionModel::mimeData(const QModelIndexList &indexes) const
{
    if( indexes.empty() ) return nullptr;
    auto mime = new QMimeData;
    mime->setData( CollectionItem::MimeType, nullptr );
    return mime;
}

//__________________________________________________________________
QVariant CollectionModel::data( const QModelIndex& index, int role ) const
{

    if( !index.isValid() ) return QVariant();

    switch( role )
    {
        case Qt::DisplayRole:
        {

            const auto& item( _find( index.internalId() ).get() );
            switch( item.type() )
            {
                case CollectionItem::Type::Artist:
                {
                    const auto& artist( item.artist() );
                    return artist.isEmpty() ? TrackInformation::UnknownArtist:artist;
                }

                case CollectionItem::Type::Album:
                {
                    auto album( item.album().isEmpty() ? TrackInformation::UnknownAlbum:item.album() );
                    if( item.disc() > 0 && item.showDiscNumber() ) album += tr(" (disc %1)" ).arg( item.disc() );
                    if( item.year() > 0 ) album += QString( " (%1)" ).arg( item.year() );
                    return album;
                }

                case CollectionItem::Type::Track:
                {
                    if( item.title().isEmpty() ) return item.titleOrFile();
                    else if( item.track() < 0 ) return item.title();
                    else return QString( "%1 %2" ).arg( QString().sprintf( "%02i", item.track() ), item.title() );
                }

                default: return QVariant();

            }
        }

        case Qt::DecorationRole:
        {

            const auto& item( _find( index.internalId() ).get() );
            switch( item.type() )
            {
                case CollectionItem::Type::Artist:
                {

                    if( showArtistIcon_ )
                    {

                        const auto icon( this->icon( item ) );
                        return icon.isNull() ? IconEngine::get( IconNames::Artist ):icon;

                    } else return IconEngine::get( IconNames::Artist );

                }

                case CollectionItem::Type::Album:
                {
                    const auto icon( this->icon( item ) );
                    return icon.isNull() ? IconEngine::get( IconNames::Album ):icon;
                }

                case CollectionItem::Type::Track: return IconEngine::get( IconNames::Track );
                default: return QVariant();
            }

        }

        default: return QVariant();
    }

}

//__________________________________________________________________
QVariant CollectionModel::headerData(int section, Qt::Orientation orientation, int role) const
{

    switch( role )
    {
        case Qt::DisplayRole:
        {
            switch( section )
            {
                case Name: return tr( "Name" );
                default: return QVariant();
            }
            break;
        }

        default: return QVariant();
    }

    // return empty
    return QVariant();

}

//__________________________________________________________________
bool CollectionModel::hasChildren( const QModelIndex& parent ) const
{
    if( !parent.isValid() ) return TreeModel::hasChildren( parent );

    const auto item( get( parent ) );
    return item.type() == CollectionItem::Type::Artist || item.type() == CollectionItem::Type::Album;
}

//__________________________________________________________________
QIcon CollectionModel::icon( const CollectionItem& item ) const
{
    switch( item.type() )
    {
        case CollectionItem::Type::Artist:
        case CollectionItem::Type::Album:
        return cache_ ? cache_->icon( item ):QIcon();

        default: return QIcon();

    }
}

//__________________________________________________________________
bool CollectionModel::updateArtist( const CollectionItem& other )
{
    if( other.type() != CollectionItem::Type::Artist ) return false;
    auto item( _root().find( other ) );
    if( item )
    {

        item->get().setTrackCount( other.trackCount() );
        item->get().setDuration( other.duration() );
        return true;

    } else return false;

}

//__________________________________________________________________
bool CollectionModel::updateAlbum( const CollectionItem& other )
{
    if( other.type() != CollectionItem::Type::Album ) return false;
    auto item( _root().find( other ) );
    if( item )
    {

        item->get().setTrackCount( other.trackCount() );
        item->get().setDuration( other.duration() );
        return true;

    } else return false;

}

//__________________________________________________________________
void CollectionModel::setCoverCache( const CoverCache::Pointer& value )
{
    cache_ = value;
    connect( cache_.get(), SIGNAL(modified()), SIGNAL(layoutAboutToBeChanged()) );
    connect( cache_.get(), SIGNAL(modified()), SIGNAL(layoutChanged()) );
}

//__________________________________________________________________
void CollectionModel::_updateConfiguration()
{
    const bool showArtistIcon = XmlOptions::get().get<bool>( "SHOW_ARTIST_ICON" );
    if( showArtistIcon_ != showArtistIcon )
    {
        emit layoutAboutToBeChanged();
        showArtistIcon_ = showArtistIcon;
        emit layoutChanged();
    }
}
