/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "FileSystemWidget.h"

#include "BaseFileInfo.h"
#include "CppUtil.h"
#include "IconEngine.h"
#include "IconNames.h"
#include "TreeView.h"
#include "XmlOptions.h"


//_____________________________________________
FileSystemWidget::FileSystemWidget( QWidget *parent ):
    BaseFileSystemWidget( parent )
{
    Debug::Throw( "FileSystemWidget::FileSystemWidget.\n" );
    _installActions();
    _updateActions();
}

//_____________________________________________
void FileSystemWidget::_addSelectionToPlaylist()
{ emit requestAddToPlaylist( _selectedFiles() ); }

//_____________________________________________
void FileSystemWidget::_replacePlaylist()
{ emit requestReplacePlaylist( _selectedFiles() ); }

//________________________________________________________
void FileSystemWidget::_updateActions()
{
    BaseFileSystemWidget::_updateActions();
    const auto selectedIndexes( _list().selectionModel()->selectedRows() );
    const bool hasSelection( !selectedIndexes.empty() );
    addToPlaylistAction_->setEnabled( hasSelection );
    replacePlaylistAction_->setEnabled( hasSelection );

    #if WITH_TAGLIB
    editTagsAction_->setEnabled( hasSelection );
    #endif

}

//________________________________________________________
File::List FileSystemWidget::_selectedFiles() const
{
    Debug::Throw( "FileSystemWidget::_selectedFiles.\n" );
    const auto records = _model().get( _list().selectionModel()->selectedRows() );
    if( records.empty() ) return File::List();
    else {

        File::List files;
        for( const auto& record:records )
        {
            // get matching file
            const auto& file( record.file() );
            if( !file.exists() ) continue;

            // in case of directory list content, recursively
            if( file.isDirectory() ) files.append( file.listFiles( static_cast<File::ListFlags>(File::Recursive|File::FollowLinks) ) );
            else files.append( file );
        }

        return files;

    }
}

//________________________________________________________
#if WITH_TAGLIB
void FileSystemWidget::_editTags()
{ emit requestEditTags( _selectedFiles() ); }
#endif

//________________________________________________________
void FileSystemWidget::_installActions()
{
    Debug::Throw( "FileSystemWidget::_installActions.\n" );

    addAction( addToPlaylistAction_ = new QAction( IconEngine::get( IconNames::Add ), tr( "Add to Playlist" ), this ) );
    connect( addToPlaylistAction_, SIGNAL(triggered()), SLOT(_addSelectionToPlaylist()) );
    addToPlaylistAction_->setEnabled(false);

    addAction( replacePlaylistAction_ = new QAction( tr( "Replace Playlist" ), this ) );
    connect( replacePlaylistAction_, SIGNAL(triggered()), SLOT(_replacePlaylist()) );
    replacePlaylistAction_->setEnabled(false);

    auto createSeparator = [&](){
        auto action = new QAction(this);
        action->setSeparator(true);
        return action; };

    auto actionList = Base::makeT<QList<QAction*>>(
    {
        addToPlaylistAction_,
        replacePlaylistAction_,
        createSeparator()
    });


    #if WITH_TAGLIB
    addAction( editTagsAction_ = new QAction( IconEngine::get( IconNames::Edit ), tr( "Edit Track Details..." ), this ) );
    connect( editTagsAction_, SIGNAL(triggered()), SLOT(_editTags()) );
    actionList.append( Base::makeT<QList<QAction*>>(
    {
        editTagsAction_,
        createSeparator()
    }));
    editTagsAction_->setEnabled(false);
    #endif

    // insert
    auto action( _contextMenu().actions().front() );
    _contextMenu().insertActions( action, actionList );

    // hide open action
    _openAction().setVisible(false);

}
