/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "LyricsProvider.h"
#include "BaseLyricsProvider_p.h"

#include "CppUtil.h"
#include "Debug.h"
#include "QUrlUtil.h"

#include <QTextCodec>
#include <QTextDocument>
#include <QNetworkRequest>
#include <QUrl>

//_____________________________________________________
LyricsProvider::LyricsProvider( QObject* parent ):
    BaseLyricsProvider( parent )
{ Debug::Throw( "LyricsProvider::LyricsProvider.\n" ); }

//_____________________________________________________
LyricsProvider::LyricsProvider( QObject* parent, const QDomElement& element ):
    LyricsProvider( parent )
{

    // parse attributes
    const auto attributes( element.attributes() );
    for( int i=0; i<attributes.count(); i++ )
    {
        const auto attribute( attributes.item( i ).toAttr() );
        if( attribute.isNull() ) continue;
        const auto name( attribute.name() );
        const auto value( attribute.value() );

        if( name == QLatin1String( "name" ) ) setName( value );
        else if( name == QLatin1String( "url" ) ) url_ = value;
        else if( name == QLatin1String( "charset" ) ) charset_ = value;
        else if( name == QLatin1String( "format" ) ) isHtml_ = ( value == QLatin1String( "html" ) );
    }

    // parse children
    // parse children elements
    for( auto&& childNode = element.firstChild(); !childNode.isNull(); childNode = childNode.nextSibling() )
    {
        const auto childElement = childNode.toElement();
        if( childElement.isNull() ) continue;

        const auto tagName( childElement.tagName() );
        if( tagName == QLatin1String( "extract" ) ) _addExtractionTags( childElement );
        else if( tagName == QLatin1String( "exclude" ) ) _addExclusionTags( childElement );
        else if( tagName == QLatin1String( "urlFormat" ) ) _addConversion( childElement );
        else if( tagName == QLatin1String( "invalidIndicator" ) ) _addInvalidIndicator( childElement );
    }

}

//_____________________________________________________
void LyricsProvider::get( const TrackInformation& trackInformation )
{

    Debug::Throw() << "LyricsProvider::get (" << name() << ")" << endl;

    // check network access manager
    if( !_hasNetworkAccessManager() )
    {
        emit lyricsNotFound( trackInformation );
        return;
    }

    // check codec
    if( !QTextCodec::codecForName(charset_.toLatin1().constData()) )
    {
        Debug::Throw(0) << "LyricsProvider::get - invalid charset: " << charset_ << endl;
        emit lyricsNotFound( trackInformation );
        return;
    }

    // create url
    const QUrl url( _replaceTags( url_, trackInformation ) );
    Debug::Throw() << "LyricsProvider::get (" << name() << ") - url: " << url.toString() << endl;

    // send request
    auto request = QNetworkRequest(url);
    auto lyricsRequest = new Private::LyricsRequest( trackInformation, _networkAccessManager().get( request ) );
    connect( lyricsRequest, SIGNAL(lyricsFound(TrackInformation,QByteArray)), SLOT(_processLyrics(TrackInformation,QByteArray)));
    connect( lyricsRequest, SIGNAL(lyricsNotFound(TrackInformation)), SIGNAL(lyricsNotFound(TrackInformation)));
    connect( lyricsRequest, SIGNAL(finished()), lyricsRequest, SLOT(deleteLater()) );

}

//_____________________________________________________
void LyricsProvider::_processLyrics( const TrackInformation& trackInformation, const QByteArray& data )
{

    Debug::Throw() << "LyricsProvider::_processLyrics (" << name() << ")" << endl;
    Debug::Throw() << "LyricsProvider::_processLyrics (" << name() << ") - data: " << data << endl;

    // get coded from charset
    const QTextCodec *codec = QTextCodec::codecForName(charset_.toLatin1().constData());
    if( !codec )
    {
        emit lyricsNotFound( trackInformation );
        return;
    }

    // get string using proper codec
    const auto dataString = codec->toUnicode( data );
    if( dataString.isEmpty() )
    {
        emit lyricsNotFound( trackInformation );
        return;
    }

    // check against invalid indicators
    for( const auto &indicator:invalidIndicatorList_ )
    {
        if( dataString.contains( indicator ) )
        {
            emit lyricsNotFound( trackInformation );
            return;
        }
    }

    // apply extraction tags
    QString lyrics;
    for( const auto extractionTag:extractionTags_ )
    {
        if( !( lyrics = _extract( dataString, extractionTag.first, extractionTag.second ) ).isEmpty() )
        { break; }
    }

    // no tags found
    if( lyrics.isEmpty() )
    {
        emit lyricsNotFound( trackInformation );
        return;
    }

    // apply exclusion tags
    for( const auto exclusionTag:exclusionTags_ )
    { lyrics = _exclude( lyrics, exclusionTag.first, exclusionTag.second ); }

    // signal
    if( lyrics.isEmpty() ) emit lyricsNotFound( trackInformation );
    else {

        Debug::Throw() << "LyricsProvider::_processLyrics - lyrics: " << lyrics << endl;

        if( isHtml_ )
        {
            QTextDocument document;
            document.setHtml( lyrics );
            lyrics = document.toPlainText();
        }

        // construct lyrics information
        LyricsInformation lyricsInformation;
        lyricsInformation.setArtist( trackInformation.artist() );
        lyricsInformation.setTitle( trackInformation.titleOrFile() );
        lyricsInformation.setProvider( name() );
        lyricsInformation.setUrl( _replaceTags( url_, trackInformation ) );
        lyricsInformation.setLyrics( lyrics );

        emit lyricsFound( trackInformation, lyricsInformation );

    }

}

//_____________________________________________________
void LyricsProvider::_addConversion( const QDomElement& element )
{
    // check element name
    if( element.tagName() != QLatin1String( "urlFormat" ) ) return;

    QRegExp regExp;
    QString replacement;

    // parse attributes
    const auto attributes( element.attributes() );
    for( int i=0; i<attributes.count(); i++ )
    {
        const auto attribute( attributes.item( i ).toAttr() );
        if( attribute.isNull() ) continue;
        const auto name( attribute.name() );
        const auto value( attribute.value() );

        if( name == QLatin1String( "replace" ) ) regExp.setPattern( QString( "[%1]" ).arg( QRegExp::escape( value ) ) );
        else if( name == QLatin1String( "with" ) ) replacement = value;
    }

    if( !regExp.pattern().isEmpty() ) addConversion( regExp, replacement );
}

//_____________________________________________________
void LyricsProvider::_addExtractionTags( const QDomElement& element )
{
    for( auto&& childNode = element.firstChild(); !childNode.isNull(); childNode = childNode.nextSibling() )
    {
        const auto childElement = childNode.toElement();
        if( childElement.isNull() ) continue;
        if( childElement.tagName() != QLatin1String( "item" ) ) continue;

        QString begin;
        QString end;

        const auto attributes( childElement.attributes() );
        for( int i=0; i<attributes.count(); i++ )
        {
            const auto attribute( attributes.item( i ).toAttr() );
            if( attribute.isNull() ) continue;
            const auto name( attribute.name() );
            const auto value( attribute.value() );
            if( name == QLatin1String( "begin" ) ) begin = value;
            else if( name == QLatin1String( "tag" ) ) begin = value;
            else if( name == QLatin1String( "end" ) ) end = value;
        }

        if( begin.isEmpty() ) continue;
        if( end.isEmpty() )
        {
            // interpret begin as an xml tag and generate end tag
            QRegExp regExp( "<(\\w+).*>");
            if( begin.indexOf( regExp ) < 0 ) continue;

            // assign end of tag to end flag
            end = QString( "</%1>" ).arg( regExp.cap(1) );
        }

        addExtractionTags( begin, end );

    }

}

//_____________________________________________________
void LyricsProvider::_addExclusionTags( const QDomElement& element )
{
    for( auto&& childNode = element.firstChild(); !childNode.isNull(); childNode = childNode.nextSibling() )
    {
        const auto childElement = childNode.toElement();
        if( childElement.isNull() ) continue;
        if( childElement.tagName() != QLatin1String( "item" ) ) continue;

        QString begin;
        QString end;

        const auto attributes( childElement.attributes() );
        for( int i=0; i<attributes.count(); i++ )
        {
            const auto attribute( attributes.item( i ).toAttr() );
            if( attribute.isNull() ) continue;
            const auto name( attribute.name() );
            const auto value( attribute.value() );
            if( name == QLatin1String( "begin" ) ) begin = value;
            else if( name == QLatin1String( "tag" ) ) begin = value;
            else if( name == QLatin1String( "end" ) ) end = value;
        }

        if( begin.isEmpty() ) continue;
        if( end.isEmpty() )
        {
            // interpret begin as an xml tag and generate end tag
            QRegExp regExp( "<(\\w+).*>");
            if( begin.indexOf( regExp ) < 0 ) continue;

            // assign end of tag to end flag
            end = QString( "</%1>" ).arg( regExp.cap(1) );
        }

        addExclusionTags( begin, end );
    }
}


//_____________________________________________________
void LyricsProvider::_addInvalidIndicator( const QDomElement& element )
{

    // parse attributes
    const auto attributes( element.attributes() );
    for( int i=0; i<attributes.count(); i++ )
    {
        const auto attribute( attributes.item( i ).toAttr() );
        if( attribute.isNull() ) continue;
        const auto name( attribute.name() );
        const auto value( attribute.value() );
        if( name == QLatin1String( "value" ) ) invalidIndicatorList_.append( value );
    }


}

//______________________________________________________________________
QString LyricsProvider::_replaceTags( const QString& data, const TrackInformation& trackInformation ) const
{
    auto out( data );
    out.replace( QLatin1String( "{Artist}" ), _convert( trackInformation.artist() ) );
    out.replace( QLatin1String( "{artist}" ), _convert( trackInformation.artist().toLower() ) );
    out.replace( QLatin1String( "{a}" ), trackInformation.artist().at(0).toLower() );
    out.replace( QLatin1String( "{Title}" ), _convert( trackInformation.title() ) );
    out.replace( QLatin1String( "{title}" ), _convert( trackInformation.title().toLower() ) );
    return out;
}


//_____________________________________________________
QString LyricsProvider::_convert( const QString& source ) const
{

    // copy source string
    auto out( source );

    // apply conversions
    for( auto conversion:conversionList_ )
    { out.replace( conversion.first, conversion.second ); }

    return out;

}
