/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "LyricsWidget.h"

#include "IconEngine.h"
#include "IconNames.h"
#include "QtUtil.h"
#include "TextEditor.h"


#include <QLayout>

//________________________________________________________
LyricsWidget::LyricsWidget( QWidget* parent ):
    QWidget( parent ),
    Counter( "LyricsWidget" )
{
    Debug::Throw( "LyricsWidget::LyricsWidget.\n" );

    _installActions();

    auto vLayout = new QVBoxLayout;
    vLayout->setMargin(0);
    vLayout->setSpacing(2);
    setLayout( vLayout );

    vLayout->addWidget( editor_ = new TextEditor( this ) );
    editor_->setWrapFromOptions( false );
    editor_->setReadOnly( true );
    editor_->setLineWrapMode( BaseEditor::WidgetWidth );

    // default formats
    const QTextCursor cursor( editor_->textCursor() );
    defaultBlockFormat_ = QTextBlockFormat( cursor.blockFormat() );
    defaultCharFormat_ = QTextCharFormat( cursor.charFormat() );

    // bold
    boldCharFormat_ = defaultCharFormat_;
    boldCharFormat_.setFontWeight( QFont::Bold );

    // italic
    italicCharFormat_ = defaultCharFormat_;
    italicCharFormat_.setFontItalic( true );

    // write header (no song)
    _writeHeader();

}

//________________________________________________________
void LyricsWidget::setTrackInformation( const TrackInformation& trackInformation )
{

    Debug::Throw( "LyricsWidget::setTrackInformation.\n" );

    // compare with current
    if( trackInformation.hashKey() == trackInformation_.hashKey() ) return;

    // assign
    trackInformation_ = trackInformation;

    // if title is invalid, try assign file instead
    if( trackInformation_.title().isEmpty() )
    { trackInformation_.setTitle( trackInformation_.titleOrFile() ); }

    // write editor header
    _writeHeader();

    // check validity
    if( _trackInformationIsValid() )
    {

        updateAction_->setEnabled( true );
        if( isVisible() ) emit requestLyrics( trackInformation_ );

    } else updateAction_->setEnabled( false );

}

//________________________________________________________
void LyricsWidget::lyricsFound( const TrackInformation& trackInformation, const LyricsInformation& lyricsInformation )
{
    Debug::Throw( "LyricsWidget::lyricsFound.\n" );

    // compare track information with current
    if( trackInformation.hashKey() != trackInformation_.hashKey() ) return;

    _writeHeader();
    if( lyricsInformation.isValid() )
    {

        _writeLyrics( lyricsInformation );
        _writeProvider( lyricsInformation );

    } else _writeNoLyrics();

}

//________________________________________________________
void LyricsWidget::lyricsNotFound( const TrackInformation& trackInformation )
{
    Debug::Throw( "LyricsWidget::lyricsNotFound.\n" );

    // compare track information with current
    if( trackInformation.hashKey() != trackInformation_.hashKey() ) return;

    _writeHeader();
    _writeNoLyrics();

}

//________________________________________________________
void LyricsWidget::requestLyrics()
{
    Debug::Throw( "LyricsWidget::requestLyrics.\n" );
    if( _trackInformationIsValid() ) emit requestLyrics( trackInformation_ );
}

//________________________________________________________
void LyricsWidget::_installActions()
{
    Debug::Throw( "LyricsWidget::_installActions.\n" );

    addAction( updateAction_ = new QAction( IconEngine::get( IconNames::Reload ), tr( "Update Collection" ), this ) );
    updateAction_->setShortcut( QKeySequence::Refresh );
    updateAction_->setShortcutContext( Qt::WidgetWithChildrenShortcut );
    connect( updateAction_, SIGNAL(triggered()), SLOT(requestLyrics()) );

}

//________________________________________________________
bool LyricsWidget::_trackInformationIsValid() const
{ return !( trackInformation_.artist().isEmpty() || trackInformation_.title().isEmpty() ); }

//________________________________________________________
void LyricsWidget::_writeHeader()
{
    Debug::Throw( "LyricsWidget::_writeHeader.\n" );
    editor_->BaseEditor::clear();

    // margin
    const int margin( editor_->fontMetrics().height()/2 );

    // move cursor to end of document
    QTextCursor cursor( editor_->textCursor() );
    cursor.movePosition( QTextCursor::End );

    // restore default and insert new block
    if( !editor_->document()->isEmpty() )
    {
        cursor.setCharFormat( defaultCharFormat_ );
        cursor.insertBlock();
    }

    // block format
    QTextBlockFormat blockFormat;
    blockFormat.setTopMargin( margin );
    blockFormat.setBottomMargin( margin );
    cursor.setBlockFormat( blockFormat );

    const auto buffer = _trackInformationIsValid() ?
        QString( "%1 - %2" ).arg( trackInformation_.artist(), trackInformation_.title() ):
        tr( "No song playing" );

    cursor.insertText( buffer, boldCharFormat_ );

    cursor.setBlockFormat( defaultBlockFormat_ );
    cursor.setCharFormat( defaultCharFormat_ );
    cursor.insertBlock();

}

//________________________________________________________
void LyricsWidget::_writeLyrics( const LyricsInformation& lyricsInformation )
{
    Debug::Throw( "LyricsWidget::_writeLyrics.\n" );
    const auto& lyrics( lyricsInformation.lyrics() );
    if( !lyrics.startsWith( '\n' ) ) editor_->insertPlainText( "\n" );
    editor_->insertPlainText( lyrics );
    if( !lyrics.endsWith( '\n' ) )  editor_->insertPlainText( "\n" );
}

//________________________________________________________
void LyricsWidget::_writeNoLyrics()
{
    Debug::Throw( "LyricsWidget::_writeNoLyrics.\n" );
    editor_->insertPlainText( tr( "\nLyrics not found." ) );
}

//________________________________________________________
void LyricsWidget::_writeProvider( const LyricsInformation& lyricsInformation )
{
    Debug::Throw( "LyricsWidget::_writeProvider.\n" );

    const auto& provider( lyricsInformation.provider() );
    if( provider.isEmpty() ) return;

    // move cursor to end of document
    QTextCursor cursor( editor_->textCursor() );
    cursor.movePosition( QTextCursor::End );

    // restore default and insert new block
    if( !editor_->document()->isEmpty() )
    {
        cursor.setCharFormat( defaultCharFormat_ );
        cursor.insertBlock();
    }

    // block format
    cursor.insertText( tr( "Downloaded from \"%1\"" ).arg( provider ), italicCharFormat_ );

    // restore default char format
    cursor.setCharFormat( defaultCharFormat_ );
    cursor.insertBlock();

}
