/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "MainWindow.h"

#include "Application.h"
#include "AudioDeviceWidget.h"
#include "BaseStatusBar.h"
#include "CollectionFindWidget.h"
#include "CollectionItem.h"
#include "CollectionView.h"
#include "CollectionWidget.h"
#include "CppUtil.h"
#include "CurrentTrackWidget.h"
#include "DeleteTracksDialog.h"
#include "FileSystemWidget.h"
#include "FileDialog.h"
#include "IconEngine.h"
#include "IconNames.h"
#include "LyricsWidget.h"
#include "LogDialog.h"
#include "MenuBar.h"
#include "MpdInformationDialog.h"
#include "MpdStatus.h"
#include "PlaybackControlToolBar.h"
#include "PlaylistParser.h"
#include "PlaylistWidget.h"
#include "PlaylistWriter.h"
#include "ReconnectionWidget.h"
#include "SavePlaylistDialog.h"
#include "SidePanelToolBar.h"
#include "SidePanelWidget.h"
#include "SingleTimeDialog.h"
#include "Singleton.h"
#include "StoredPlaylistView.h"
#include "StoredPlaylistWidget.h"
#include "TrackInformation.h"
#include "TreeView.h"
#include "Util.h"
#include "VolumeControlMenu.h"
#include "WarningDialog.h"

#if WITH_TAGLIB
#include "TagEditionDialog.h"
#endif

#if WITH_DBUS
#include "DBusCoreAdaptor.h"
#include "DBusPlayerAdaptor.h"
#include <QDBusConnection>
#endif

#include <QLayout>
#include <QNetworkAccessManager>
#include <QSplitter>
#include <QTimer>

#include <algorithm>
#include <numeric>

namespace
{
    namespace Local
    {

        #if WITH_TAGLIB
        //___________________________________________________________
        //* transform collection item list into file list, with absolute path names
        File::List fileList( const CollectionItem::List& items )
        {

            File musicDirectory( XmlOptions::get().raw( "MUSIC_DIR" ) );
            musicDirectory.addTrailingSlash();

            File::List files;
            std::transform( items.begin(), items.end(), std::back_inserter( files ),
                [&musicDirectory]( const CollectionItem& item )
            {
                auto file( item.file() );
                return file.addPath( musicDirectory );
            });

            return files;

        }
        #endif

        //___________________________________________________________
        QString findArtistCommandString( const TrackInformation& item )
        { return QString( "find artist \"%1\"" ).arg( item.artist() ); }

        //___________________________________________________________
        QString findAlbumCommandString( const TrackInformation& item )
        {
            auto commandString = QString( "find artist \"%1\" album \"%2\"" ).arg( item.artist(), item.album() );
            if( item.disc() > 0 ) commandString += QString( " disc \"%1\"" ).arg( item.disc() );
            return commandString;
        }

        //___________________________________________________________
        QString findTrackCommandString( const TrackInformation& item )
        {
            auto commandString = QString( "find artist \"%1\" album \"%2\"" ).arg( item.artist(), item.album() );
            if( item.disc() > 0 ) commandString += QString( " disc \"%1\"" ).arg( item.disc() );
            if( item.track() > 0 ) commandString += QString( " track \"%1\"" ).arg( item.track() );
            else if( !item.title().isEmpty() ) commandString += QString( " title \"%1\"" ).arg( item.title() );
            return commandString;
        }

        //___________________________________________________________
        Mpd::Command findArtistCommand( const TrackInformation& item )
        { return Mpd::Command( Mpd::Command::Type::FindTracks, findArtistCommandString( item ) ); }

        //___________________________________________________________
        Mpd::Command findAlbumCommand( const TrackInformation& item )
        { return Mpd::Command( Mpd::Command::Type::FindTracks, findAlbumCommandString( item ) ); }

        //___________________________________________________________
        Mpd::Command findTrackCommand( const TrackInformation& item )
        { return Mpd::Command( Mpd::Command::Type::FindTracks, findTrackCommandString( item ) ); }

        //___________________________________________________________
        Mpd::Command findPlaylistTracksCommand( const StoredPlaylistItem& item )
        {
            return Mpd::Command(
                Mpd::Command::Type::StoredPlaylistInfo,
                QString( "listplaylistinfo \"%1\"" ).arg( item.playlist() ) );
        }
    }

}

//_________________________________________________________
MainWindow::MainWindow( QWidget* parent ):
    BaseMainWindow( parent ),
    Counter( "MainWindow" )
{

    setOptionName( "MAIN_WINDOW" );

    #if WITH_TAGLIB
    // tag edition dialog
    tagEditionDialog_ = new TagEditionDialog( this );
    #endif

    logDialog_ = new LogDialog( this );

    // install actions
    _installActions();

    // add close action from Application, to make sure the shortcut works
    auto application = Base::Singleton::get().application<Application>();
    addAction( &application->closeAction() );

    // control toolbar
    playbackControlToolBar_ = new PlaybackControlToolBar( this );
    playbackControlToolBar_->setTransparent( true );
    connect( playbackControlToolBar_, SIGNAL(requestStartPlayback()), SLOT(_startPlayback()) );
    connect( playbackControlToolBar_, SIGNAL(requestResumePlayback()), SLOT(_resumePlayback()) );
    connect( playbackControlToolBar_, SIGNAL(requestPausePlayback()), SLOT(pausePlayback()) );
    connect( playbackControlToolBar_, SIGNAL(requestStopPlayback()), SLOT(stopPlayback()) );
    connect( playbackControlToolBar_, SIGNAL(requestPreviousTrack()), SLOT(_previousTrack()) );
    connect( playbackControlToolBar_, SIGNAL(requestNextTrack()), SLOT(_nextTrack()) );

    connect( playbackControlToolBar_, SIGNAL(requestRepeatNone()), SLOT(_repeatNone()) );
    connect( playbackControlToolBar_, SIGNAL(requestRepeatOne()), SLOT(_repeatOne()) );
    connect( playbackControlToolBar_, SIGNAL(requestRepeatAll()), SLOT(_repeatAll()) );
    connect( playbackControlToolBar_, SIGNAL(requestRandomPlayback(bool)), SLOT(_random(bool)) );

    connect( &playbackControlToolBar_->volumeControlMenu(), SIGNAL(volumeChanged(int)), SLOT(_setVolume(int)) );
    connect( &playbackControlToolBar_->currentTrackWidget(), SIGNAL(requestSeek(qreal)), SLOT(_seek(qreal)) );

    // main widget
    mainWidget_ = new QWidget( this );
    setCentralWidget( mainWidget_ );

    auto vLayout = new QVBoxLayout;
    vLayout->setMargin(0);
    vLayout->setSpacing(0);
    mainWidget_->setLayout( vLayout );

    // reconnection widget
    vLayout->addWidget( reconnectionWidget_ = new ReconnectionWidget( this ), 0 );
    reconnectionWidget_->hide();

    // splitter
    auto splitter( new QSplitter( mainWidget_ ) );
    splitter->setOrientation( Qt::Horizontal );
    vLayout->addWidget( splitter, 1 );

    // collection
    sidePanelWidget_ = new SidePanelWidget;
    sidePanelWidget_->setDefaultWidth( XmlOptions::get().get<int>( "SIDE_PANEL_WIDTH" ) );
    splitter->addWidget( sidePanelWidget_ );

    connect( &sidePanelWidget_->visibilityAction(), SIGNAL(toggled(bool)), SLOT(_toggleSidePanelWidget(bool)) );

    connect( &sidePanelWidget_->collectionWidget(), SIGNAL(requestCollectionUpdate()), SLOT(updateCollection()) );
    connect( &sidePanelWidget_->collectionWidget(), SIGNAL(requestAlbums(CollectionItem)), SLOT(_listAlbums(CollectionItem)) );
    connect( &sidePanelWidget_->collectionWidget(), SIGNAL(requestTracks(CollectionItem)), SLOT(_listTracks(CollectionItem)) );
    connect( &sidePanelWidget_->collectionWidget(), SIGNAL(requestAddToPlaylist(CollectionItem::List)), SLOT(_addFromCollection(CollectionItem::List)) );
    connect( &sidePanelWidget_->collectionWidget(), SIGNAL(requestReplacePlaylist(CollectionItem::List)), SLOT(_replaceFromCollection(CollectionItem::List)) );
    connect( &sidePanelWidget_->collectionWidget(), SIGNAL(requestDeleteTracks(CollectionItem::List)), SLOT(_deleteTracksFromCollection(CollectionItem::List)) );
    connect( &sidePanelWidget_->collectionWidget().findWidget(), SIGNAL(requestSearchCollection(QString,TrackInformation::Mask)), SLOT(_searchCollection(QString,TrackInformation::Mask)) );

    connect( &sidePanelWidget_->storedPlaylistWidget(), SIGNAL(requestTracks(StoredPlaylistItem)), SLOT(_listPlaylistTracks(StoredPlaylistItem)) );
    connect( &sidePanelWidget_->storedPlaylistWidget(), SIGNAL(requestAddToPlaylist(StoredPlaylistItem::List)), SLOT(_addFromStoredPlaylist(StoredPlaylistItem::List)) );
    connect( &sidePanelWidget_->storedPlaylistWidget(), SIGNAL(requestReplacePlaylist(StoredPlaylistItem::List)), SLOT(_replaceFromStoredPlaylist(StoredPlaylistItem::List)) );
    connect( &sidePanelWidget_->storedPlaylistWidget(), SIGNAL(requestDeleteItems(StoredPlaylistItem::List)), SLOT(_removeFromStoredPlaylist(StoredPlaylistItem::List)) );
    connect( &sidePanelWidget_->storedPlaylistWidget(), SIGNAL(requestRenamePlaylist(StoredPlaylistItem,QString)), SLOT(_renameStoredPlaylist(StoredPlaylistItem,QString)) );
    connect( &sidePanelWidget_->storedPlaylistWidget(), SIGNAL(requestExportPlaylist(StoredPlaylistItem)), SLOT(_exportStoredPlaylist(StoredPlaylistItem)) );
    connect( &sidePanelWidget_->storedPlaylistWidget().model(), SIGNAL(requestAddToPlaylist(StoredPlaylistItem,int)), SLOT(_addToStoredPlaylist(StoredPlaylistItem,int)) );
    connect( &sidePanelWidget_->storedPlaylistWidget().model(), SIGNAL(requestMoveToPlaylist(StoredPlaylistItem,int)), SLOT(_moveToStoredPlaylist(StoredPlaylistItem,int)) );

    connect( &sidePanelWidget_->fileSystemWidget(), SIGNAL(requestAddToPlaylist(File::List)), SLOT(addFromFileSystem(File::List)) );
    connect( &sidePanelWidget_->fileSystemWidget(), SIGNAL(requestReplacePlaylist(File::List)), SLOT(replaceFromFileSystem(File::List)) );

    // playlist
    splitter->addWidget( playlistWidget_ = new PlaylistWidget );
    connect( &playlistWidget_->removeTrackAction(), SIGNAL(triggered()), SLOT(_removeSelectedTracks()) );
    connect( &playlistWidget_->clearPlaylistAction(), SIGNAL(triggered()), SLOT(clearPlaylist()) );
    connect( &playlistWidget_->savePlaylistAction(), SIGNAL(triggered()), SLOT(_savePlaylist()) );
    connect( &playlistWidget_->list(), SIGNAL(activated(QModelIndex)), SLOT(_startPlayback()) );
    connect( &playlistWidget_->model(), SIGNAL(requestAddFiles(File::List,int)), SLOT(_addFiles(File::List,int)) );
    connect( &playlistWidget_->model(), SIGNAL(requestMoveTracks(int)), SLOT(_moveTracks(int)) );
    connect( &playlistWidget_->model(), SIGNAL(requestAddFromCollection(int)), SLOT(_addFromCollection(int)) );
    connect( &playlistWidget_->model(), SIGNAL(requestAddFromStoredPlaylist(int)), SLOT(_addFromStoredPlaylist(int)) );

    #if WITH_CDIO_PARANOIA
    connect( &sidePanelWidget_->audioDeviceWidget(), SIGNAL(requestAddToPlaylist(TrackInformation::List)), SLOT(_addFromAudioDevice(TrackInformation::List)) );
    connect( &sidePanelWidget_->audioDeviceWidget(), SIGNAL(requestReplacePlaylist(TrackInformation::List)), SLOT(_replaceFromAudioDevice(TrackInformation::List)) );
    connect( &playlistWidget_->model(), SIGNAL(requestAddFromAudioDevice(int)), SLOT(_addFromAudioDevice(int)) );
    #endif

    #if WITH_TAGLIB
    connect( &sidePanelWidget_->collectionWidget(), SIGNAL(requestEditTags(CollectionItem::List)), SLOT(_editTags(CollectionItem::List)) );
    connect( &sidePanelWidget_->fileSystemWidget(), SIGNAL(requestEditTags(File::List)), SLOT(_editTags(File::List)) );
    connect( playlistWidget_, SIGNAL( requestEditTags(File::List)), SLOT(_editTags(File::List)) );
    #endif

    // assign stretch factors
    splitter->setStretchFactor( 0, 0 );
    splitter->setStretchFactor( 1, 1 );
    connect( splitter, SIGNAL(splitterMoved(int,int)), SLOT(_splitterMoved()) );

    // menubar
    setMenuBar( new MenuBar( this ) );
    menuBar()->installEventFilter( this );

    // navigation toolbar
    sidePanelToolBar_ = new SidePanelToolBar( this );
    sidePanelToolBar_->connect( *sidePanelWidget_ );

    // configuration
    connect( application, SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    _updateStatusBarVisibility();

}

//_________________________________________________________
CurrentTrackWidget& MainWindow::currentTrackWidget() const
{ return playbackControlToolBar_->currentTrackWidget(); }

//_________________________________________________________
VolumeControlMenu& MainWindow::volumeControlMenu() const
{ return playbackControlToolBar_->volumeControlMenu(); }

//_________________________________________________________
CollectionWidget& MainWindow::collectionWidget() const
{ return sidePanelWidget_->collectionWidget(); }

//_________________________________________________________
LyricsWidget& MainWindow::lyricsWidget() const
{ return sidePanelWidget_->lyricsWidget(); }

//_________________________________________________________
StoredPlaylistWidget& MainWindow::storedPlaylistWidget() const
{ return sidePanelWidget_->storedPlaylistWidget(); }

//_________________________________________________________
#if WITH_CDIO_PARANOIA
AudioDeviceWidget& MainWindow::audioDeviceWidget() const
{ return sidePanelWidget_->audioDeviceWidget(); }
#endif

//_________________________________________________________
bool MainWindow::eventFilter( QObject* object, QEvent* event )
{

    // check object, option and event type
    if( object != menuBar() ) return BaseMainWindow::eventFilter( object, event );
    if( !( XmlOptions::get().get<bool>( "COLLAPSE_ON_DOUBLE_CLICK" ) && event->type() == QEvent::MouseButtonDblClick ) )
    { return BaseMainWindow::eventFilter( object, event ); }

    // check buttons and modifiers
    auto mouseEvent = static_cast<QMouseEvent*>( event );
    if( !( mouseEvent->button() == Qt::LeftButton && mouseEvent->modifiers() == Qt::NoModifier ) )
    { return BaseMainWindow::eventFilter( object, event ); }

    event->accept();
    _toggleCollapse();
    return true;

}

//_________________________________________________________
QMenu* MainWindow::createPopupMenu()
{

    Debug::Throw( "MainWindow::createPopupMenu.\n" );
    auto menu( BaseMainWindow::createPopupMenu() );

    if( menu->actions().isEmpty() ) menu->addAction( collapseAction_ );
    else
    {
        auto action = menu->insertSeparator( menu->actions().front() );
        menu->insertAction( action, collapseAction_ );
    }

    return menu;
}

//_________________________________________________________
void MainWindow::clear()
{
    Debug::Throw( "MainWindow::clear.\n" );
    _setState( Mpd::State::Stopped );
    playlistWidget_->model().clear();
    sidePanelWidget_->collectionWidget().model().clear();
    playlistWidget_->statusBar().label(0).clear();
}

//_________________________________________________________
void MainWindow::mpdInformation()
{
    Debug::Throw( "MainWindow::mpdInformation.\n" );

    if( !mpdInformationDialog_ ) mpdInformationDialog_ = new Mpd::InformationDialog( this );

    mpdInformationDialog_->clear();
    mpdInformationDialog_->setMpdVersion( mpdVersion_ );

    emit commandRequested(
        Mpd::Command( Mpd::Command::Type::Statistics,
        { "stats", "urlhandlers", "decoders", "tagtypes" } ) );

    mpdInformationDialog_->exec();

}

//_________________________________________________________
void MainWindow::updateStatus()
{
    Debug::Throw( "MainWindow::updateStatus.\n" );
    emit commandRequested( Mpd::Command( Mpd::Command::Type::Status, "status" ) );
}

//_________________________________________________________
void MainWindow::clearPlaylist()
{
    Debug::Throw( "MainWindow::_clearPlaylist.\n" );
    if( !playlistWidget_->model().empty() )
    { emit commandRequested( Mpd::Command( Mpd::Command::Type::ClearPlaylist, "clear" ) ); }
}

//_________________________________________________________
void MainWindow::updatePlaylist()
{
    Debug::Throw( "MainWindow::updatePlaylist\n" );
    emit commandRequested( Mpd::Command( Mpd::Command::Type::PlaylistInfo, "playlistinfo" ) );
}

//_________________________________________________________
void MainWindow::updateCollection()
{
    Debug::Throw( "MainWindow::updateCollection.\n" );
    emit commandRequested( Mpd::Command( Mpd::Command::Type::UpdateCollection, "update" ) );
}

//_________________________________________________________
void MainWindow::updateCollection( QString path )
{
    Debug::Throw() << "MainWindow::updateCollection - path: " << path << endl;
    emit commandRequested( Mpd::Command( Mpd::Command::Type::UpdateCollection, QString( "update \"%1\"" ).arg( path ) ) );
}

//_________________________________________________________
void MainWindow::loadCollection()
{
    Debug::Throw( "MainWindow::loadCollection\n" );

    if( searchActive_ )
    {

        _searchCollection( searchedString_, searchMask_ );

    } else {

        /*
        count group artist command is used for the collection widget, total time and tag edition dialog
        the empty artist count is used to load tracks that have no artist
        list album command is used for the tag edition dialog and update number of albums per artist
        */
        auto commandStringList = Base::makeT<QStringList>( { "count group artist", "count artist \"\"", "list album group artist" } );
        for( const auto& item:sidePanelWidget_->collectionWidget().expandedItems() )
        {
            if( item.type() == CollectionItem::Type::Artist )
            { commandStringList.append( QString( "find artist \"%1\"" ).arg( item.artist() ) ); }
        }

        emit commandRequested( Mpd::Command( Mpd::Command::Type::LoadCollection, commandStringList ) );

    }

}

//_________________________________________________________
void MainWindow::loadStoredPlaylists()
{
    Debug::Throw( "MainWindow::loadStoredPlaylists\n" );
    QStringList commandStringList( "listplaylists" );
    for( const auto& item:sidePanelWidget_->storedPlaylistWidget().expandedItems() )
    {
        if( item.type() == StoredPlaylistItem::Type::Playlist )
        { commandStringList.append( QString( "listplaylistinfo \"%1\"" ).arg( item.playlist() ) ); }
    }

    emit commandRequested( Mpd::Command( Mpd::Command::Type::LoadPlaylists, commandStringList ) );
}

//_________________________________________________________
void MainWindow::addFromFileSystem( const File::List& files )
{ _addFiles( files, playlistWidget_->model().rowCount() ); }

//_________________________________________________________
void MainWindow::replaceFromFileSystem( const File::List& files )
{
    // clear model
    playlistWidget_->model().clear();

    // emit clear command
    emit commandRequested( Mpd::Command( Mpd::Command::Type::ClearPlaylist, "clear" ) );

    // add to playlist
    addFromFileSystem( files );

}

//_________________________________________________________
void MainWindow::pausePlayback()
{
    Debug::Throw( "MainWindow::pausePlayback.\n" );
    playbackControlToolBar_->currentTrackWidget().stopTimer();
    _setState( Mpd::State::Paused );
    _updatePlaybackControlButtons();
    emit commandRequested( Mpd::Command( Mpd::Command::Type::Pause, "pause 1" ) );
}

//_________________________________________________________
void MainWindow::stopPlayback()
{
    Debug::Throw( "MainWindow::stopPlayback.\n" );
    playbackControlToolBar_->currentTrackWidget().stopTimer();
    _setState( Mpd::State::Stopped );
    _updatePlaybackControlButtons();
    emit commandRequested( Mpd::Command( Mpd::Command::Type::Stop, "stop" ) );
}

//_________________________________________________________
void MainWindow::processCommandReply( const Mpd::Command& command )
{
    Debug::Throw( "MainWindow::processCommandReply.\n" );

    // split
    const auto commands = command.subCommands();
    if( commands.empty() ) return;

    switch( command.type() )
    {
        case Mpd::Command::Type::LoadCollection: return _processLoadCollectionCommand( commands );
        case Mpd::Command::Type::LoadPlaylists: return _processLoadPlaylistsCommand( commands );

        case Mpd::Command::Type::Status:
        for( const auto& command:commands )
        { if( command.success() ) _processStatusCommand( command ); }
        break;

        case Mpd::Command::Type::Statistics:
        for( const auto& command:commands )
        { if( command.success() ) _processMpdInformationCommand( command ); }
        break;

        case Mpd::Command::Type::ClearPlaylist:
        for( const auto& command:commands )
        { if( command.success() ) _processClearCommand( command ); }
        break;

        case Mpd::Command::Type::PlaylistInfo:
        for( const auto& command:commands )
        { if( command.success() ) _processPlaylistInfoCommand( command ); }
        break;

        case Mpd::Command::Type::FindTracks:
        for( const auto& command:commands )
        { if( command.success() ) _processFindTracksCommand( command ); }
        break;

        case Mpd::Command::Type::FindTrack:
        for( const auto& command:commands )
        { if( command.success() ) _processFindTrackCommand( command ); }
        break;

        case Mpd::Command::Type::SearchCollection:
        for( const auto& command:commands )
        { if( command.success() ) _processCollectionSearchCommand( command ); }
        break;

        case Mpd::Command::Type::UpdateCollection:
        for( const auto& command:commands )
        { if( command.success() ) _processCollectionUpdateCommand( command ); }
        break;

        case Mpd::Command::Type::StoredPlaylistInfo:
        for( const auto& command:commands )
        { if( command.success() ) _processStoredPlaylistInfoCommand( command ); }
        break;

        default: break;

    }

}

//____________________________________________
void MainWindow::updateCover( const TrackInformation&, const TrackInformation& track )
{

    CollectionItem item( std::move( track ) );
    item.setType( CollectionItem::Type::Album );

    if( !collectionItemKeys_.contains( item.hashKey() ) )
    {
        emit requestCover( item );
        collectionItemKeys_.insert( item.hashKey() );
    }

}

//____________________________________________
void MainWindow::closeEvent( QCloseEvent* event )
{

    Debug::Throw( "MainWindow::closeEvent.\n" );

    // accept event
    event->accept();

    // trigger close action, unless systemtray icon is used
    if( !( QSystemTrayIcon::isSystemTrayAvailable() && XmlOptions::get().get<bool>( "USE_SYSTEMTRAY" ) ) )
    {

        Base::Singleton::get().application<Application>()->closeAction().trigger();

    } else if( !XmlOptions::get().contains( "SHOW_CLOSE_DIALOG" ) || XmlOptions::get().get<bool>( "SHOW_CLOSE_DIALOG" ) ) {

        SingleTimeDialog dialog( this, tr(
            "Closing the main window will keep Musicbox running in the System Tray. "
            "Use <b>Exit</b> from the Musicbox tray icon to exit the application." ) );

        dialog.setWindowTitle( tr( "System Tray" ) );
        dialog.exec();
        XmlOptions::get().set<bool>( "SHOW_CLOSE_DIALOG", !dialog.dontShowAgain() );

    }

}

//_______________________________________________________
void MainWindow::timerEvent( QTimerEvent* event )
{

    if( event->timerId() == resizeTimer_.timerId() )
    {

        // stop timer
        resizeTimer_.stop();

        // save size
        if( sidePanelWidget_->visibilityAction().isChecked() )
        { XmlOptions::get().set<int>( "SIDE_PANEL_WIDTH", sidePanelWidget_->width() ); }

    } else if( event->timerId() == collectionUpdateTimer_.timerId() ) {

        loadCollection();

    } else return BaseMainWindow::timerEvent( event );

}

//________________________________________________________
void MainWindow::mouseDoubleClickEvent( QMouseEvent* event )
{

    // check button and modifiers
    if( !( XmlOptions::get().get<bool>( "COLLAPSE_ON_DOUBLE_CLICK" ) && event->button() == Qt::LeftButton && event->modifiers() == Qt::NoModifier ) )
    { return BaseMainWindow::mouseDoubleClickEvent( event ); }

    event->accept();
    _toggleCollapse();

}

//________________________________________________________
void MainWindow::_updateConfiguration()
{
    Debug::Throw( "MainWindow::_updateConfiguration.\n" );
    sidePanelWidget_->visibilityAction().setChecked( XmlOptions::get().get<bool>("SHOW_SIDE_PANEL_WIDGET") );
    toggleLogAction_->setChecked( XmlOptions::get().get<bool>( "LOG_ENABLED" ) );

    #if WITH_DBUS
    bool dbusEnabled( XmlOptions::get().get<bool>( "USE_DBUS_INTERFACE" ) );
    if( dbusEnabled_ != dbusEnabled )
    {
        if( ( dbusEnabled_ = dbusEnabled ) ) QTimer::singleShot( 0, this, SLOT(_initDBus()) );
        else _deinitDBus();
    }
    #endif

    // collapsed state
    if( XmlOptions::get().get<bool>( "COLLAPSED" ) )
    {

        if( XmlOptions::get().contains( "EXPANDED_SIZE_WIDTH" ) &&  XmlOptions::get().contains( "EXPANDED_SIZE_HEIGHT" ) )
        { expandedSize_ = QSize( XmlOptions::get().get<int>( "EXPANDED_SIZE_WIDTH" ), XmlOptions::get().get<int>( "EXPANDED_SIZE_HEIGHT" ) ); }

        _setCollapsed( true );

    } else {

        if( XmlOptions::get().contains( "COLLAPSED_SIZE_WIDTH" ) &&  XmlOptions::get().contains( "COLLAPSED_SIZE_HEIGHT" ) )
        { collapsedSize_ = QSize( XmlOptions::get().get<int>( "COLLAPSED_SIZE_WIDTH" ), XmlOptions::get().get<int>( "COLLAPSED_SIZE_HEIGHT" ) ); }

        _setCollapsed( false );

    }

}

//________________________________________________________
void MainWindow::_toggleSidePanelWidget( bool state )
{
    Debug::Throw( "MainWindow::_toggleSidePanelWidget.\n" );
    XmlOptions::get().set<bool>( "SHOW_SIDE_PANEL_WIDGET", state );
}

//________________________________________________________
void MainWindow::_splitterMoved()
{
    Debug::Throw( "MainWindow::_splitterMoved.\n" );
    resizeTimer_.start( 200, this );
}

//_________________________________________________________
void MainWindow::_addFiles( const File::List& files, int row )
{
    Debug::Throw( "MainWindow::_addFiles.\n" );

    if( files.empty() ) return;
    else if( files.size() == 1 && PlaylistParser::isPlaylist( files.front() ) )
    {
        QStringList commandStringList;
        Mpd::Command::List commandList;
        const auto trackInformationList( PlaylistParser( files.front() ).parse() );
        for( const auto& trackInformation:trackInformationList )
        {
            if( trackInformation.file().exists() ) commandStringList.append( QString( "addid \"file://%1\" %2" ).arg( trackInformation.file().get() ).arg( row++ ) );
            else if( !( trackInformation.artist().isEmpty() || trackInformation.album().isEmpty() || trackInformation.title().isEmpty() ) ) {

                auto command( Local::findTrackCommand( trackInformation ) );
                command.setOptionList( { "row", QString::number( row ) } );
                commandList.prepend( command );

            }
        }

        if( !commandStringList.isEmpty() )
        { emit commandRequested(  Mpd::Command( Mpd::Command::Type::AddTracks, commandStringList ) ); }

        if( !commandList.isEmpty() ) emit commandsRequested( commandList );

    } else {

        Mpd::Command::List commands;
        std::transform( files.begin(), files.end(), std::front_inserter( commands ),
            [&row]( const File& file )
            { return Mpd::Command( Mpd::Command::Type::AddTracks, QString( "addid \"file://%1\" %2" ).arg( file ).arg( row ) ); } );
        if( !commands.empty() ) emit commandsRequested( commands );
    }

}

//_________________________________________________________
void MainWindow::_moveTracks( int row )
{
    Debug::Throw() << "MainWindow::_moveTracks - row: " << row << endl;

    // get selection from playlist model
    const auto selectedIndexes( playlistWidget_->list().selectionModel()->selectedRows() );
    if( selectedIndexes.empty() ) return;
    int firstTrack( selectedIndexes.front().row() );
    int lastTrack( selectedIndexes.back().row() + 1 );

    // adjust row
    if( row >= lastTrack ) row -= lastTrack-firstTrack;

    emit commandRequested( Mpd::Command( Mpd::Command::Type::MoveTracks, QString( "move %1:%2 %3" ).arg( firstTrack ).arg( lastTrack ).arg( row ) ) );

}

//_________________________________________________________
void MainWindow::_addFromCollection( int row )
{

    Debug::Throw( "MainWindow::_addFromCollection.\n" );

    // get index list
    auto indexList( sidePanelWidget_->collectionWidget().list().selectionModel()->selectedRows() );

    // remove indexes for which parent is already in list
    indexList.erase(
        std::remove_if( indexList.begin(), indexList.end(),
        [&indexList]( const QModelIndex& index )
        { return indexList.contains( index.parent() ); } ),
        indexList.end() );

    const auto items( sidePanelWidget_->collectionWidget().model().get( indexList ) );
    _doAddFromCollection( items, row );

}

//_________________________________________________________
void MainWindow::_addFromCollection( const CollectionItem::List& constItems )
{

    Debug::Throw( "MainWindow::_addFromCollection.\n" );

    // remove items for which parent is already in list
    auto items( constItems );
    items.erase(
        std::remove_if( items.begin(), items.end(), [&items]( const CollectionItem& item )
        {
            CollectionItem parent( item );
            parent.setType( item.parentType() );
            return items.contains( parent );
        } ), items.end() );

    // row is set to last from playlist
    _doAddFromCollection( items, playlistWidget_->model().rowCount() );
}

//_________________________________________________________
void MainWindow::_replaceFromCollection( const CollectionItem::List& items )
{

    Debug::Throw( "MainWindow::_replaceFromCollection.\n" );

    // clear model
    playlistWidget_->model().clear();

    // emit clear command
    emit commandRequested( Mpd::Command( Mpd::Command::Type::ClearPlaylist, "clear" ) );

    // add to playlist
    _addFromCollection( items );

}

//_________________________________________________________
void MainWindow::_deleteTracksFromCollection( const CollectionItem::List& constItems )
{

    Debug::Throw( "MainWindow::_deleteTracksFromCollection.\n" );

    // create delete tracks dialog if not exist
    if( !deleteTracksDialog_ ) deleteTracksDialog_ = new DeleteTracksDialog( this );

    // remove items for which parent is already in list
    auto items( constItems );
    items.erase(
        std::remove_if( items.begin(), items.end(), [&items]( const CollectionItem& item )
        {
            CollectionItem parent( item );
            parent.setType( item.parentType() );
            return items.contains( parent );
        } ), items.end() );

    // assign to dialog
    deleteTracksDialog_->setItems( items );

    // generate necessary find commands
    QStringList commandStringList;
    for( const auto& item:items )
    {
        switch( item.type() )
        {
            default: continue;
            case CollectionItem::Type::Artist:
            commandStringList.append( Local::findArtistCommandString( item ) );
            break;

            case CollectionItem::Type::Album:
            commandStringList.append( Local::findAlbumCommandString( item ) );
            break;
        }
    }

    if( !commandStringList.isEmpty() )
    {
        Mpd::Command command( Mpd::Command::Type::FindTracks, commandStringList );
        command.addOption( "delete" );
        emit commandRequested( command );
    }

    // run dialog and delete
    if( deleteTracksDialog_->exec() ) deleteTracksDialog_->deleteTracks();

}

//_________________________________________________________
void MainWindow::_searchCollection( const QString& searchedString, TrackInformation::Mask mask )
{
    Debug::Throw( "MainWindow::_searchCollection.\n" );

    // store mask
    searchMask_ = mask;
    searchedString_ = searchedString;

    // process
    if( searchedString.isEmpty() || mask == TrackInformation::None )
    {

        if( searchActive_ )
        {
            searchActive_ = false;
            loadCollection();
        }

    } else {

        searchActive_ = true;
        if( mask == TrackInformation::Artist ) emit commandRequested( Mpd::Command( Mpd::Command::Type::SearchCollection, QString( "search artist \"%1\"" ).arg( searchedString ) ) );
        else if( mask == TrackInformation::Title ) emit commandRequested( Mpd::Command( Mpd::Command::Type::SearchCollection,QString( "search title \"%1\"" ).arg( searchedString ) ) );
        else emit commandRequested( Mpd::Command( Mpd::Command::Type::SearchCollection, QString( "search any \"%1\"" ).arg( searchedString ) ) );

    }

}

//_________________________________________________________
void MainWindow::_listPlaylistTracks( const StoredPlaylistItem& item )
{
    Debug::Throw( "MainWindow::_listPlaylistTracks.\n" );
    if( item.type() == StoredPlaylistItem::Type::Playlist && !item.playlist().isEmpty() )
    { emit commandRequested( Local::findPlaylistTracksCommand( item ) ); }
}

//_________________________________________________________
void MainWindow::_addFromStoredPlaylist( int row )
{

    Debug::Throw( "MainWindow::_addFromCollection.\n" );

    // get index list
    auto indexList( sidePanelWidget_->storedPlaylistWidget().list().selectionModel()->selectedRows() );

    // remove indexes for which parent is already in list
    indexList.erase(
        std::remove_if( indexList.begin(), indexList.end(),
        [&indexList]( const QModelIndex& index )
        { return indexList.contains( index.parent() ); } ),
        indexList.end() );

    const auto items( sidePanelWidget_->storedPlaylistWidget().model().get( indexList ) );
    _doAddFromStoredPlaylist( items, row );

}

//_________________________________________________________
void MainWindow::_addFromStoredPlaylist( const StoredPlaylistItem::List& constItems )
{

    Debug::Throw( "MainWindow::_addFromStoredPlaylist.\n" );

    // remove items for which parent is already in list
    auto items( constItems );
    items.erase(
        std::remove_if( items.begin(), items.end(), [&items]( const StoredPlaylistItem& item )
        {
            StoredPlaylistItem parent( item );
            parent.setType( item.parentType() );
            return items.contains( parent );
        } ), items.end() );

    // row is set to last from playlist
    _doAddFromStoredPlaylist( items, playlistWidget_->model().rowCount() );
}

//_________________________________________________________
void MainWindow::_replaceFromStoredPlaylist( const StoredPlaylistItem::List& items )
{

    Debug::Throw( "MainWindow::_replaceFromStoredPlaylist.\n" );

    // clear model
    playlistWidget_->model().clear();

    // emit clear command
    emit commandRequested( Mpd::Command( Mpd::Command::Type::ClearPlaylist, "clear" ) );

    // add to playlist
    _addFromStoredPlaylist( items );

}

//_________________________________________________________
void MainWindow::_removeFromStoredPlaylist( const StoredPlaylistItem::List& constItems )
{

    Debug::Throw( "MainWindow::_removeFromStoredPlaylist.\n" );

    // remove items for which parent is already in list
    auto items( constItems );
    items.erase(
        std::remove_if( items.begin(), items.end(), [&items]( const StoredPlaylistItem& item )
        {
            StoredPlaylistItem parent( item );
            parent.setType( item.parentType() );
            return items.contains( parent );
        } ), items.end() );

    QStringList commandStringList;
    std::transform( items.begin(), items.end(), std::front_inserter( commandStringList ),
        []( const StoredPlaylistItem& item )
        {
            switch( item.type() )
            {
                default:
                case StoredPlaylistItem::Type::Playlist:
                return QString( "rm \"%1\"" ).arg( item.playlist() );

                case StoredPlaylistItem::Type::Track:
                return QString( "playlistdelete \"%1\" %2" ).arg( item.playlist() ).arg( item.trackIndex() );
            }
        });

    if( !commandStringList.isEmpty() )
    { emit commandRequested( Mpd::Command( Mpd::Command::Type::RemoveTracks, commandStringList ) ); }

}

//_________________________________________________________
void MainWindow::_addToStoredPlaylist( const StoredPlaylistItem& parentItem, int row )
{

    Debug::Throw( "MainWindow::_addToStoredPlaylist.\n" );

    if( parentItem.type() != StoredPlaylistItem::Type::Playlist ) return;

    const auto items( playlistWidget_->model().get( playlistWidget_->list().selectionModel()->selectedRows() ) );
    if( items.isEmpty() ) return;

    QStringList commandStringList;
    std::transform( items.begin(), items.end(), std::back_inserter( commandStringList ),
        [&parentItem]( const TrackInformation& track )
        { return QString( "playlistadd \"%1\" \"%2\"" ).arg( parentItem.playlist(), track.file().get() ); } );

    if( row >= 0 && row < parentItem.trackCount() )
    {
        int previousRow( parentItem.trackCount() );
        std::transform( items.begin(), items.end(), std::back_inserter( commandStringList ),
            [&parentItem, &previousRow, &row]( const TrackInformation& )
            { return QString( "playlistmove \"%1\" %2 %3" ).arg( parentItem.playlist() ).arg( previousRow++ ).arg( row++ ); } );
    }

    emit commandRequested( Mpd::Command( Mpd::Command::Type::SavePlaylist, commandStringList ) );

}


//_________________________________________________________
void MainWindow::_moveToStoredPlaylist( const StoredPlaylistItem& parentItem, int row )
{

    Debug::Throw( "MainWindow::_addToStoredPlaylist.\n" );

    if( parentItem.type() != StoredPlaylistItem::Type::Playlist ) return;

    const auto storedPlaylistWidget( &sidePanelWidget_->storedPlaylistWidget() );
    const auto items( storedPlaylistWidget->model().get( storedPlaylistWidget->list().selectionModel()->selectedRows() ) );
    if( items.isEmpty() ) return;

    // check items
    if( std::any_of( items.begin(), items.end(), StoredPlaylistItem::DifferentTypeFTor( StoredPlaylistItem::Type::Track ) ) )
    { return; }

    int firstTrack = items.front().trackIndex();
    int trackCount = parentItem.trackCount();
    int indexOffset = 0;

    QStringList commandStringList;
    for( const auto& item:items )
    {
        if( item.isChild( parentItem ) )
        {

            if( row < 0 ) row = trackCount;
            if( row <= firstTrack ) commandStringList.append( QString( "playlistmove \"%1\" %2 %3" ).arg( parentItem.playlist() ).arg( firstTrack++ ).arg( row++ ) );
            else commandStringList.append( QString( "playlistmove \"%1\" %2 %3" ).arg( parentItem.playlist() ).arg( firstTrack ).arg( row-1 ) );

        } else {

            commandStringList.append( QString( "playlistdelete \"%1\" %2" ).arg( item.playlist() ).arg( item.trackIndex()-(indexOffset++) ) );
            commandStringList.append( QString( "playlistadd \"%1\" \"%2\"" ).arg( parentItem.playlist(), item.file().get() ) );

            if( row >= 0 && row < trackCount )
            { commandStringList.append( QString( "playlistmove \"%1\" %2 %3" ).arg( parentItem.playlist() ).arg( trackCount++ ).arg( row++ ) ); }

        }

    }

    emit commandRequested( Mpd::Command( Mpd::Command::Type::SavePlaylist, commandStringList ) );

}

//_________________________________________________________
void MainWindow::_renameStoredPlaylist( const StoredPlaylistItem& item, const QString& playlist )
{

    Debug::Throw( "MainWindow::_renameStoredPlaylist.\n" );
    if(
        item.type() != StoredPlaylistItem::Type::Playlist ||
        playlist.isEmpty() ||
        item.playlist() == playlist )
        { return; }

    emit commandRequested(
        Mpd::Command( Mpd::Command::Type::SavePlaylist,
        QString( "rename \"%1\" \"%2\"" ).arg( item.playlist(), playlist ) ) );

}

//_________________________________________________________
void MainWindow::_exportStoredPlaylist( const StoredPlaylistItem& item )
{

    Debug::Throw( 0, "MainWindow::_exportStoredPlaylist.\n" );
    if( item.type() != StoredPlaylistItem::Type::Playlist ) return;

    // create command
    auto command( Local::findPlaylistTracksCommand( item ) );
    command.addOption( "export" );

    // request
    emit commandRequested( command );

}

#if WITH_CDIO_PARANOIA
//_________________________________________________________
void MainWindow::_addFromAudioDevice( int row )
{

    Debug::Throw( "MainWindow::_addFromAudioDevice.\n" );

    // get index list
    auto indexList( sidePanelWidget_->audioDeviceWidget().list().selectionModel()->selectedRows() );
    const auto tracks( sidePanelWidget_->audioDeviceWidget().model().get( indexList ) );

    // add
    _doAddFromAudioDevice( tracks, row );

}

//_________________________________________________________
void MainWindow::_addFromAudioDevice( const TrackInformation::List& tracks )
{ _doAddFromAudioDevice( tracks, playlistWidget_->model().rowCount() ); }

//_________________________________________________________
void MainWindow::_replaceFromAudioDevice( const TrackInformation::List& tracks )
{

    Debug::Throw( "MainWindow::_replaceFromAudioDevice.\n" );

    // clear model
    playlistWidget_->model().clear();

    // emit clear command
    emit commandRequested( Mpd::Command( Mpd::Command::Type::ClearPlaylist, "clear" ) );

    // add to playlist
    _addFromAudioDevice( tracks );

}
#endif

//_________________________________________________________
void MainWindow::_removeSelectedTracks()
{
    Debug::Throw( "MainWindow::_removeSelectedTracks.\n" );
    const auto indexes( playlistWidget_->list().selectionModel()->selectedRows() );
    if( indexes.empty() ) return;
    else if( indexes.size() == 1 ) emit commandRequested( Mpd::Command( Mpd::Command::Type::RemoveTracks, QString( "delete %1" ).arg( indexes.front().row() ) ) );
    else emit commandRequested( Mpd::Command( Mpd::Command::Type::RemoveTracks, QString( "delete %1:%2" ).arg( indexes.front().row() ).arg( indexes.back().row()+1 ) ) );

    // clear selection
    playlistWidget_->list().selectionModel()->clear();

}

//_________________________________________________________
void MainWindow::_savePlaylist()
{
    Debug::Throw( "MainWindow::_savePlaylist.\n" );
    if( playlistWidget_->model().empty() ) return;
    SavePlaylistDialog dialog( this );
    if( dialog.exec() )
    {
        const auto playlist( dialog.playlist() );
        if( !playlist.isEmpty() )
        { emit commandRequested( Mpd::Command( Mpd::Command::Type::SavePlaylist, QString( "save \"%1\"" ).arg( playlist ) ) ); }
    }
}

//_________________________________________________________
void MainWindow::_startPlayback()
{
    Debug::Throw( "MainWindow::_startPlayback.\n" );

    // abort fade if any
    playbackControlToolBar_->abortFade();

    const auto& index( playlistWidget_->model().currentTrackIndex() );
    const auto& trackInformation( playlistWidget_->model().get( index ) );

    // update total and elapsed time
    playbackControlToolBar_->currentTrackWidget().setTotalTime( trackInformation.duration() );
    playbackControlToolBar_->currentTrackWidget().setElapsedTime( 0 );

    _setState( Mpd::State::Playing );
    _notifyCurrentTrack();
    _updatePlaybackControlButtons();

    emit commandRequested(
        Mpd::Command( Mpd::Command::Type::Play, QString( "play %1" )
        .arg( index.isValid() ? index.row():0 ) ) );
}

//_________________________________________________________
void MainWindow::_resumePlayback()
{
    Debug::Throw( "MainWindow::_resumePlayback.\n" );
    _setState( Mpd::State::Playing );
    _notifyCurrentTrack();
    _updatePlaybackControlButtons();
    emit commandRequested( Mpd::Command( Mpd::Command::Type::Resume, "pause 0" ) );
}

//_________________________________________________________
void MainWindow::_previousTrack()
{
    Debug::Throw( "MainWindow::_previousTrack.\n" );
    const bool isRandom( playbackControlToolBar_->isRandomPlayback() );
    if( ( isRandom || playlistWidget_->selectPreviousTrack() ) && state_ != Mpd::State::Stopped )
    {
        if( state_ == Mpd::State::Playing ) _notifyCurrentTrack();
        emit commandRequested( Mpd::Command( Mpd::Command::Type::PreviousTrack, "previous" ) );
    }

    _updatePlaybackControlButtons();
    _updateWindowTitle();

}

//_________________________________________________________
void MainWindow::_nextTrack()
{
    Debug::Throw( "MainWindow::_nextTrack.\n" );
    const bool isRandom( playbackControlToolBar_->isRandomPlayback() );
    if( (isRandom || playlistWidget_->selectNextTrack() ) && state_ != Mpd::State::Stopped )
    {
        if( state_ == Mpd::State::Playing ) _notifyCurrentTrack();
        emit commandRequested( Mpd::Command( Mpd::Command::Type::NextTrack, "next" ) );
    }

    _updatePlaybackControlButtons();
    _updateWindowTitle();
}

//_________________________________________________________
void MainWindow::_setVolume( int value )
{
    Debug::Throw( "MainWindow::_setVolume.\n" );
    emit commandRequested( Mpd::Command( Mpd::Command::Type::ChangeVolume, QString( "setvol %1" ).arg( value ) ) );
    emit statusChanged( Volume );
}

//_________________________________________________________
void MainWindow::_seek( qreal value )
{
    Debug::Throw() << "MainWindow::_seek - value: " << value << endl;
    emit commandRequested( Mpd::Command( Mpd::Command::Type::Seek, QString( "seekcur %1" ).arg( value ) ) );
}

//_________________________________________________________
void MainWindow::_repeatNone()
{
    Debug::Throw( "MainWindow::_repeatNone.\n" );
    playlistWidget_->model().setRepeatMode( Mpd::RepeatMode::RepeatNone );
    _updatePlaybackControlButtons();
    _notifyPlaybackMode();
    emit commandRequested( Mpd::Command(
        Mpd::Command::Type::ChangeRepeatMode,
        { QLatin1String( "repeat 0" ), QLatin1String( "single 0" ) } ) );
    emit statusChanged( PlaybackMode );
}

//_________________________________________________________
void MainWindow::_repeatOne()
{
    Debug::Throw( "MainWindow::_repeatOne.\n" );
    playlistWidget_->model().setRepeatMode( Mpd::RepeatMode::RepeatOne );
    _updatePlaybackControlButtons();
    _notifyPlaybackMode();
    emit commandRequested( Mpd::Command(
        Mpd::Command::Type::ChangeRepeatMode,
        { QLatin1String( "repeat 1" ), QLatin1String( "single 1" ) } ) );
    emit statusChanged( PlaybackMode );
}

//_________________________________________________________
void MainWindow::_repeatAll()
{
    Debug::Throw( "MainWindow::_repeatAll.\n" );
    playlistWidget_->model().setRepeatMode( Mpd::RepeatMode::RepeatAll );
    _updatePlaybackControlButtons();
    _notifyPlaybackMode();
    emit commandRequested( Mpd::Command(
        Mpd::Command::Type::ChangeRepeatMode,
        { QLatin1String( "repeat 1" ), QLatin1String( "single 0" ) } ) );
    emit statusChanged( PlaybackMode );
}


//_________________________________________________________
void MainWindow::_random( bool value )
{
    Debug::Throw( "MainWindow::_random.\n" );
    _updatePlaybackControlButtons();
    _notifyPlaybackMode();
    emit commandRequested( Mpd::Command( Mpd::Command::Type::ChangeRepeatMode, QString( "random %1" ).arg( value ) ) );
    emit statusChanged( PlaybackMode );
}

//_________________________________________________________
void MainWindow::_listAlbums( const CollectionItem& item )
{
    Debug::Throw( "MainWindow::_listAlbums.\n" );

    /*
    do nothing if there is an active search
    because all found items are already in the model
    */
    if( searchActive_ ) return;

    if( item.type() == CollectionItem::Type::Artist )
    { emit commandRequested( Local::findArtistCommand( item ) ); }
}

//_________________________________________________________
void MainWindow::_listTracks( const CollectionItem& item )
{
    Debug::Throw( "MainWindow::_listTracks.\n" );

    /*
    do nothing if there is an active search
    because all found items are already in the model
    */
    if( searchActive_ ) return;

    if( item.type() == CollectionItem::Type::Album )
    { emit commandRequested( Local::findAlbumCommand( item ) ); }
}

//_________________________________________________________
#if WITH_TAGLIB
void MainWindow::_editTags( const CollectionItem::List& constItems )
{

    // remove items for which parent is already in list
    auto items( constItems );
    items.erase(
        std::remove_if( items.begin(), items.end(), [&items]( const CollectionItem& item )
        {
            CollectionItem parent( item );
            parent.setType( item.parentType() );
            return items.contains( parent );
        } ), items.end() );

    if( searchActive_ )
    {

        // when search is active, we use only items that are in the model
        CollectionItem::List expandedItems;
        for( const auto& item:items )
        {
            if( item.type() == CollectionItem::Type::Track ) expandedItems.append( item );
            else {

                // copy all track children to expanded indexes
                auto children( sidePanelWidget_->collectionWidget().model().children( item ) );
                std::copy(
                    children.begin(),
                    std::remove_if( children.begin(), children.end(), CollectionItem::DifferentTypeFTor( CollectionItem::Type::Track ) ),
                    std::back_inserter( expandedItems ) );
            }
        }

        _editTags( Local::fileList( expandedItems ) );

    } else {

        // store tracks
        CollectionItem::List tracks;

        // store find commands (for adding albums or artists)
        Mpd::Command::List commandList;

        // loop over items
        for( const auto& item:items )
        {
            switch( item.type() )
            {
                case CollectionItem::Type::Track:
                tracks.append( item );
                break;

                case CollectionItem::Type::Artist:
                {
                    auto command( Local::findArtistCommand( item ) );
                    command.setOptionList( { "addTag" } );

                    // find commands are added in reverse order, in order to preserve ordering
                    commandList.append( command );
                    break;
                }

                case CollectionItem::Type::Album:
                {
                    auto command( Local::findAlbumCommand( item ) );
                    command.setOptionList( { "addTag" } );

                    // find commands are  added in reverse order, in order to preserve ordering
                    commandList.append( command );
                    break;
                }

                default: break;
            }

        }

        if( commandList.empty() ) _editTags( Local::fileList( tracks ) );
        else {

            // add files from tracks to tag edition dialog
            tagEditionDialog_->reset();
            tagEditionDialog_->addFiles( Local::fileList( tracks ) );

            // make sure last sent command triggers showning the dialog
            commandList.back().addOption( "endTagEdition" );

            // send command
            emit commandsRequested( commandList );

        }

    }

}

//_________________________________________________________
void MainWindow::_editTags( const File::List& files )
{

    // check files
    if( files.empty() ) return;

    // create dialog
    if( tagEditionDialog_->setFiles( files ) && tagEditionDialog_->exec() )
    {
        auto path( tagEditionDialog_->location() );
        if( !( path.exists() && path.isDirectory() ) ) path = path.path();

        // music directory
        File musicDirectory( XmlOptions::get().raw( "MUSIC_DIR" ) );
        musicDirectory.addTrailingSlash();

        // request collection update
        if( path.startsWith( musicDirectory ) )
        { updateCollection( path.get().mid( musicDirectory.get().size() ) ); }
    }

}
#endif

//_______________________________________________
void MainWindow::_showLog()
{
    Debug::Throw( "MainWindow::_showLog.\n" );
    logDialog_->show();
    logDialog_->toggleSticky( false );
    logDialog_->raise();
    XmlOptions::get().set<bool>( "SHOW_LOG", true );
}

//_______________________________________________
void MainWindow::_toggleLog( bool value )
{
    Debug::Throw( "MainWindow::_toggleLog.\n" );
    if( !value ) logDialog_->hide();

    logDialog_->setEnabled( value );
    showLogAction_->setVisible( value );
    XmlOptions::get().set<bool>( "LOG_ENABLED", value );

}

//_________________________________________________________
void MainWindow::_initDBus()
{
    #if WITH_DBUS
    Debug::Throw( "MainWindow::_initDBus.\n" );

    // register service
    QDBusConnection::sessionBus().registerService("org.mpris.MediaPlayer2.musicbox");

    // create adaptors
    dbusCoreAdaptor_.reset( new DBus::CoreAdaptor(this) );
    dbusPlayerAdaptor_.reset( new DBus::PlayerAdaptor(this) );

    // register object
    QDBusConnection::sessionBus().registerObject("/org/mpris/MediaPlayer2", this, QDBusConnection::ExportAdaptors);
    #endif
}

//_________________________________________________________
void MainWindow::_deinitDBus()
{
    #if WITH_DBUS
    Debug::Throw( "MainWindow::_deinitDBus.\n" );

    // unregister object
    QDBusConnection::sessionBus().unregisterObject("/org/mpris/MediaPlayer2" );

    // delete adaptors
    dbusCoreAdaptor_.reset();
    dbusPlayerAdaptor_.reset();

    // unregister service
    QDBusConnection::sessionBus().unregisterService("org.mpris.MediaPlayer2.musicbox");
    #endif
}

//_____________________________________________
void MainWindow::_toggleCollapse()
{
    Debug::Throw( "MainWindow::_toggleCollapse.\n" );

    // store sizes
    if( collapsed_ )
    {

        collapsedSize_ = size();
        XmlOptions::get().set<int>( "COLLAPSED_SIZE_WIDTH", collapsedSize_.width() );
        XmlOptions::get().set<int>( "COLLAPSED_SIZE_HEIGHT", collapsedSize_.height() );

    } else {

        expandedSize_ = size();
        XmlOptions::get().set<int>( "EXPANDED_SIZE_WIDTH", expandedSize_.width() );
        XmlOptions::get().set<int>( "EXPANDED_SIZE_HEIGHT", expandedSize_.height() );

    }

    // collapse
    _setCollapsed( !collapsed_ );
}

//_____________________________________________
void MainWindow::_setCollapsed( bool value )
{
    Debug::Throw( "MainWindow::_setCollapsed.\n" );

    // check if changed
    if( collapsed_ == value ) return;

    // store
    collapsed_ = value;
    XmlOptions::get().set<bool>( "COLLAPSED", value );

    if( value )
    {
        // calculate collapse size if not available
        if( !collapsedSize_.isValid() )
        { collapsedSize_ = playbackControlToolBar_->size(); }

        // hide widgets and resize
        menuBar()->hide();
        sidePanelToolBar_->hide();
        mainWidget_->hide();
        adjustSize();
        resize( collapsedSize_ );
        setMaximumHeight( collapsedSize_.height() );

        // update button
        collapseAction_->setText( tr( "Expand Interface" ) );

    } else {

        // show widgets and resize
        sidePanelToolBar_->show();
        mainWidget_->show();
        menuBar()->setVisible( showMenuBarAction().isChecked() );
        adjustSize();

        // reset maximum height
        setMaximumHeight( QWIDGETSIZE_MAX );

        // adjust size
        if( expandedSize_.isValid() )
        { resize( expandedSize_ ); }

        // update button
        collapseAction_->setText( tr( "Collapse Interface" ) );

    }

}

//_____________________________________________
void MainWindow::_installActions()
{

    Debug::Throw( "FileSelectionFrame::_installActions.\n" );

    // log
    showLogAction_ = new QAction( IconEngine::get( IconNames::Find ), tr( "Show Log" ), this );
    showLogAction_->setVisible( false );
    connect( showLogAction_, SIGNAL(triggered()), SLOT(_showLog()) );

    toggleLogAction_ = new QAction( tr( "Enable Log" ), this );
    toggleLogAction_->setCheckable( true );
    toggleLogAction_->setChecked( false );
    connect( toggleLogAction_, SIGNAL(toggled(bool)), SLOT(_toggleLog(bool)) );

    // collapse
    collapseAction_ = new QAction( tr( "Collapse" ), this );
    connect( collapseAction_, SIGNAL(triggered()), SLOT(_toggleCollapse()) );

}

//_________________________________________________________
void MainWindow::_setState( Mpd::State state )
{
    Debug::Throw() << "MainWindow::_setState - " << Base::toIntegralType( state ) << endl;
    state_ = state;
    playbackControlToolBar_->setState( state_ );
    playbackControlToolBar_->currentTrackWidget().setState( state_ );
    playlistWidget_->model().setState( state_ );
}

//_________________________________________________________
void MainWindow::_updateWindowTitle()
{
    Debug::Throw( "MainWindow::_updateWindowTitle.\n" );

    const auto& index( playlistWidget_->model().currentTrackIndex() );
    if( playlistWidget_->model().contains( index ) )
    {
        const auto& track( playlistWidget_->model().get( index ) );
        const auto artist( track.artist().isEmpty() ? TrackInformation::UnknownArtist:track.artist() );
        setWindowTitle( QString( "%1%2%3" ).arg( artist, Util::windowTitleSeparator(), track.titleOrFile() ) );
    } else QMainWindow::setWindowTitle( "Musicbox" );

}

//_________________________________________________________
void MainWindow::_updatePlaybackControlButtons()
{
    Debug::Throw( "MainWindow::_updatePlaybackControlButtons.\n" );
    const bool empty( playlistWidget_->model().empty() );
    playbackControlToolBar_->startPlaybackAction().setEnabled( !empty || state_ != Mpd::State::Stopped );
    playbackControlToolBar_->stopPlaybackAction().setEnabled( state_ != Mpd::State::Stopped );

    if( empty )
    {

        playbackControlToolBar_->previousTrackAction().setEnabled( false );
        playbackControlToolBar_->nextTrackAction().setEnabled( false );

    } else if( playbackControlToolBar_->isRandomPlayback() ) {

        playbackControlToolBar_->previousTrackAction().setEnabled( state_ != Mpd::State::Stopped );
        playbackControlToolBar_->nextTrackAction().setEnabled( state_ != Mpd::State::Stopped );

    } else {

        playbackControlToolBar_->previousTrackAction().setEnabled( playlistWidget_->model().hasPreviousTrack() );
        playbackControlToolBar_->nextTrackAction().setEnabled( playlistWidget_->model().hasNextTrack() );

    }

    // trigger status update
    emit statusChanged( PlaybackState|PreviousTrack|NextTrack );
}

//_________________________________________________________
void MainWindow::_notifyCurrentTrack()
{
    Debug::Throw( "MainWindow::_notifyCurrentTrack.\n" );
    const auto& index = playlistWidget_->model().currentTrackIndex();
    if( index.isValid() ) emit newNotification( playlistWidget_->model().get( index ) );
}

//_________________________________________________________
void MainWindow::_notifyPlaybackMode()
{
    Debug::Throw( "MainWindow::_notifyPlaybackMode.\n" );
    playlistWidget_->statusBar().label(1).setVisible( playbackControlToolBar_->repeatMode() != Mpd::RepeatMode::RepeatNone );
    playlistWidget_->statusBar().label(2).setVisible( playbackControlToolBar_->isRandomPlayback() );
}

//_________________________________________________________
void MainWindow::_processStatusCommand( const Mpd::Command& command )
{

    Debug::Throw() << "MainWindow::_processStatusCommand - status: " << command.reply() << endl;
    Mpd::Status status( command.reply() );

    // update state
    _setState( status.state() );

    // check update job Id
    if( status.updateJobId() > 0 )
    {

        sidePanelWidget_->collectionWidget().busy();
        if( !collectionUpdateTimer_.isActive() ) collectionUpdateTimer_.start( 2000, this );

        if( !collectionUpdateTimeStamp_.isValid() )
        {
            Debug::Throw(0, "MainWindow::_processStatusCommand - update started.\n" );
            collectionUpdateTimeStamp_ = TimeStamp::now();
        }

    } else {

        if( collectionUpdateTimer_.isActive() ) collectionUpdateTimer_.stop();
        sidePanelWidget_->collectionWidget().idle();

        if( collectionUpdateTimeStamp_.isValid() )
        {
            Debug::Throw(0)
                << "MainWindow::_processStatusCommand - update duration: "
                << TrackInformation::durationString( static_cast<quint32>(TimeStamp::now().unixTime() - collectionUpdateTimeStamp_.unixTime()) )
                << endl;

            collectionUpdateTimeStamp_ = TimeStamp();
        }

    }

    // update repeat mode
    playbackControlToolBar_->lock();
    playbackControlToolBar_->setRepeatMode( status.repeatMode() );
    playbackControlToolBar_->setRandomPlayback( status.random() );
    playbackControlToolBar_->unlock();

    playlistWidget_->model().setRepeatMode( status.repeatMode() );

    // update volume
    playbackControlToolBar_->volumeControlMenu().lock();

    if( !playbackControlToolBar_->volumeControlMenu().isMuted() )
    { playbackControlToolBar_->volumeControlMenu().setVolume( status.volume() ); }

    playbackControlToolBar_->updateVolumeIcon();
    playbackControlToolBar_->updateVolumeToolTip();
    playbackControlToolBar_->setRepeatMode( status.repeatMode() );
    playbackControlToolBar_->volumeControlMenu().unlock();

    // update current track widget
    playbackControlToolBar_->currentTrackWidget().setTotalTime( status.totalTime() );
    playbackControlToolBar_->currentTrackWidget().setElapsedTime( status.elapsedTime() );

    // stop/start timer
    if( state_ == Mpd::State::Playing ) playbackControlToolBar_->currentTrackWidget().startTimer();
    else playbackControlToolBar_->currentTrackWidget().stopTimer();

    // update playlist
    const auto currentTrackIndex( status.currentTrackIndex() );
    if( playlistWidget_->setCurrentTrackIndex( currentTrackIndex ) && currentTrackIndex >= 0 )
    {

        // get track
        const auto& currentTrack( playlistWidget_->model().currentTrack() );

        // enable seek state
        playbackControlToolBar_->currentTrackWidget().setSeekEnabled( !currentTrack.isAudioStream() );

        // forward elapsed time
        if( currentTrack.isAudioStream() ) emit elapsedTimeChanged( status.elapsedTime() );

        // emit notification
        if( state_ == Mpd::State::Playing ) _notifyCurrentTrack();

    }

    // update lyrics track information
    sidePanelWidget_->lyricsWidget().setTrackInformation( playlistWidget_->model().currentTrack() );

    // trigger status update
    emit statusChanged( Volume|CurrentTrack|PlaybackMode );

    // update window title
    _updateWindowTitle();
    _updatePlaybackControlButtons();
    _notifyPlaybackMode();

}

//_________________________________________________________
void MainWindow::_processMpdInformationCommand( const Mpd::Command& command )
{

    Debug::Throw() << "MainWindow::_processMpdInformationCommand - status: " << command.reply() << endl;
    if( mpdInformationDialog_ ) mpdInformationDialog_->update( command.reply() );

}

//_________________________________________________________
void MainWindow::_processClearCommand( const Mpd::Command& )
{
    Debug::Throw( "MainWindow::_processClearCommand.\n" );
    playlistWidget_->model().clear();
    playlistWidget_->statusBar().label(0).clear();
}

//_________________________________________________________
void MainWindow::_processPlaylistInfoCommand( const Mpd::Command& command )
{
    Debug::Throw( "MainWindow::_processPlaylistInfoCommand.\n" );

    // parse tracks
    const auto tracks( TrackInformation::Helper::list<TrackInformation>( command.reply() ) );
    if( tracks.empty() )
    {

        playlistWidget_->model().clear();
        playlistWidget_->statusBar().label(0).clear();

    } else {

        // assign to playlist
        playlistWidget_->saveState();
        playlistWidget_->model().set( tracks );
        playlistWidget_->restoreState();

        // update status bar
        const auto totalDuration = std::accumulate( tracks.begin(), tracks.end(), quint32(0),
            []( quint32 value, const TrackInformation& track )
            { return value + track.duration(); } );

        if( tracks.size() == 1 ) playlistWidget_->statusBar().label(0).setText( tr( "1 track (%1)" ).arg( TrackInformation::durationString( totalDuration, true ) ) );
        else playlistWidget_->statusBar().label(0).setText( tr( "%1 tracks (%2)" ).arg( tracks.size() ).arg( TrackInformation::durationString( totalDuration, true ) ) );

        // request covers
        for( auto&& track:tracks )
        {
            CollectionItem item( std::move( track ) );
            item.setType( CollectionItem::Type::Album );
            if( !collectionItemKeys_.contains( item.hashKey() ) )
            {
                emit requestCover( item );
                collectionItemKeys_.insert( item.hashKey() );
            }
        }

    }

    playlistWidget_->updateActions();
    _updatePlaybackControlButtons();

}

//_________________________________________________________
void MainWindow::_processStoredPlaylistInfoCommand( const Mpd::Command& command )
{
    Debug::Throw( "MainWindow::_processStoredPlaylistInfoCommand.\n" );
    const auto& commandString( command.commandList().front() );

    // capture playlist
    QRegExp regExp( "listplaylistinfo \"(.*)\"");
    if( commandString.indexOf( regExp ) < 0 ) return;
    const auto playlist = regExp.cap(1);

    // read tracks
    auto items( TrackInformation::Helper::list<StoredPlaylistItem>( command.reply() ) );

    // assign playlist, type and index
    int index = 0;
    for( auto& item:items )
    {
        item.setType( StoredPlaylistItem::Type::Track );
        item.setPlaylist( playlist );
        item.setTrackIndex( index++ );
    }

    if( !items.empty() )
    {

        // lock playlists widget and save expanded items
        sidePanelWidget_->storedPlaylistWidget().lock();
        sidePanelWidget_->storedPlaylistWidget().saveExpandedItems();

        std::sort( items.begin(), items.end() );
        sidePanelWidget_->storedPlaylistWidget().updatePlaylists( items );

        // restore expanded items and unlock
        sidePanelWidget_->storedPlaylistWidget().restoreExpandedItems();
        sidePanelWidget_->storedPlaylistWidget().unlock();

        // also update covers
        CollectionItem::List collectionItems;
        std::transform( items.begin(), items.end(), std::back_inserter( collectionItems ),
            []( const StoredPlaylistItem& item ) { return CollectionItem( item ); } );
        std::sort( collectionItems.begin(), collectionItems.end() );
        _updateCovers( collectionItems );

    }

    // check option to see if needs insertion in playlist
    int position( command.optionList().indexOf( "row" ) );
    if( position >= 0 )
    {

        // make sure items are sorted
        std::sort( items.begin(), items.end() );

        // get row from arguments
        int row = command.optionList()[position+1].toInt();

        QStringList commandStringList;
        std::transform( items.begin(), items.end(), std::front_inserter( commandStringList ),
            [&row]( const StoredPlaylistItem& item )
            { return QString( "addid \"%1\" %2" ).arg( item.file() ).arg( row ); } );

        emit commandRequested( Mpd::Command( Mpd::Command::Type::AddTracks, commandStringList ) );

    }

    // check option for exporting playlist
    if( command.optionList().contains( "export" ) )
    {

        // get default file name
        auto file = File( QString( "%1.%2" ).arg( StoredPlaylistItem::displayName(playlist), PlaylistWriter::defaultExtension() ) )
            .addPath( File( XmlOptions::get().raw( "MUSIC_DIR" ) ) );

        FileDialog dialog( this );
        dialog.setAcceptMode( QFileDialog::AcceptSave );
        dialog.setFileMode( QFileDialog::AnyFile );
        dialog.setFilter( "*.m3u *.xspf;;All files (*)" );
        dialog.setCaption( tr( "Export Playlist" ) );
        dialog.selectFile( file );

        // get file from dialog
        file = dialog.getFile();
        if( file.isEmpty() ) return;

        QFile qfile( file.get() );
        if( !qfile.open( QIODevice::WriteOnly ) )
        {
            WarningDialog( this, tr( "Failed to write playlist to file \"%1\"" ).arg( file.localName() ) ).exec();
            return;
        }

        qfile.write( PlaylistWriter().create( items, PlaylistWriter::playlistFormat( file ) ) );
        qfile.close();
    }

}

//_________________________________________________________
void MainWindow::_processCollectionUpdateCommand( const Mpd::Command& command )
{
    Debug::Throw() << "MainWindow::_processCollectionUpdateCommand - command: " << command.commandList().join( "\n" ) << endl;

    // parse output
    QRegExp regExp( "^updating_db:\\s+(\\d+)" );
    if( command.reply().indexOf( regExp ) >= 0 )
    { sidePanelWidget_->collectionWidget().busy(); }

}

//_________________________________________________________
void MainWindow::_processLoadCollectionCommand( const Mpd::Command::List& commands )
{
    Debug::Throw( "MainWindow::_processLoadCollectionCommand.\n" );

    if( commands.empty() ) return;

    // keep track of artists, albums and tracks
    CollectionItem::List artists;
    CollectionItem::List albums;
    CollectionItem::List tracks;

    QStringList artistList;
    QStringList albumList;

    // loop over commands
    for( const auto& command:commands )
    {

        if( !command.success() ) continue;

        const auto& commandString( command.commandList().front() );
        if( commandString.startsWith( "count " ) )
        {

            QRegExp commandRegExp( "artist\\s+\"(.*)\"" );
            commandRegExp.setMinimal(true);
            if( commandString.indexOf( commandRegExp ) >= 0 )
            {

                // get artist
                const auto artist = commandRegExp.cap(1);

                // get reply
                const auto& reply( command.reply() );
                QRegExp regExp(  "playtime:\\s+(\\d+)\\n" );
                int duration = 0;
                if( reply.indexOf( regExp ) >= 0 && (duration = regExp.cap(1).toInt()) > 0 )
                {
                    auto item( CollectionItem( CollectionItem::Type::Artist, artist ) );
                    item.setDuration( duration );
                    artists.append( item );
                }

            } else {

                // get reply
                const auto& reply( command.reply() );

                QRegExp regExp(  "Artist:\\s+(.*)\\n.*\\nplaytime:\\s+(\\d+)\\n" );
                regExp.setMinimal( true );

                for( int position = 0; ( position = reply.indexOf( regExp, position ) ) >= 0; position += regExp.matchedLength() )
                {
                    const auto artist = regExp.cap(1);
                    if( !artist.isEmpty() ) artistList.append( artist );
                    auto item( CollectionItem( CollectionItem::Type::Artist, artist ) );
                    item.setDuration( regExp.cap(2).toInt() );
                    artists.append( item );
                }

            }

        } else if( commandString.startsWith( "list " ) ) {

            const auto arguments( commandString.split( ' ', QString::SkipEmptyParts ) );
            if( arguments.size() < 2 ) continue;

            // get reply
            const auto& reply( command.reply() );

            if( arguments[1] == "artist" )
            {

                QRegExp regExp( "Artist:\\s+(.*)\\n" );
                regExp.setMinimal( true );

                for( int position = 0; ( position = reply.indexOf( regExp, position ) ) >= 0; position += regExp.matchedLength() )
                {
                    const auto artist = regExp.cap(1);
                    if( !artist.isEmpty() ) artistList.append( artist );
                    artists.append( CollectionItem( CollectionItem::Type::Artist, artist ) );
                }

            } else if( arguments[1] == "album" && arguments.size() > 3 && arguments[2] == "group" && arguments[3] == "artist" ) {

                QRegExp regExp( "Album:\\s+(.*)\\nArtist:\\s+(.*)\\n" );
                regExp.setMinimal( true );

                for( int position = 0; ( position = reply.indexOf( regExp, position ) ) >= 0; position += regExp.matchedLength() )
                {
                    const auto album = regExp.cap(1);
                    const auto artist = regExp.cap(2);
                    albums.append( CollectionItem( CollectionItem::Type::Album, artist, album ) );

                    if( !album.isEmpty() ) albumList.append( album );

                }

            }

        } else if( commandString.startsWith( "find " ) ) {

            tracks.append( TrackInformation::Helper::list<CollectionItem>( command.reply() ) );

        }

    }

    // lock collection widget and save expanded items
    sidePanelWidget_->collectionWidget().lock();
    sidePanelWidget_->collectionWidget().saveExpandedItems();

    if( !artists.empty() )
    {
        // sort artists
        std::sort( artists.begin(), artists.end() );

        // assign number of album counts
        std::sort( albums.begin(), albums.end() );
        for( auto& artist:artists )
        {
            auto iter = std::partition( albums.begin(), albums.end(), TrackInformation::SameArtistFTor( artist ) );
            artist.setAlbumCount( iter - albums.begin() );
            albums.erase( albums.begin(), iter );
        }

        // assign to model
        sidePanelWidget_->collectionWidget().model().set( artists );

        // request covers
        for( const auto& artist:artists )
        {
            // emit cover request if necessary
            if( !collectionItemKeys_.contains( artist.hashKey() ) )
            {
                emit requestCover( artist );
                collectionItemKeys_.insert( artist.hashKey() );
            }

        }
    }

    if( !tracks.empty() )
    {
        // split items by artist and assign albums to model
        std::sort( tracks.begin(), tracks.end() );
        sidePanelWidget_->collectionWidget().updateArtists( tracks );

        // split items by album and assign tracks to model
        sidePanelWidget_->collectionWidget().updateAlbums( tracks );

        // update covers
        _updateCovers( tracks );

    }

    // restore expanded items and unlock
    sidePanelWidget_->collectionWidget().restoreExpandedItems();
    sidePanelWidget_->collectionWidget().unlock();

    #if WITH_TAGLIB
    // assign artists and albums to tag edition widget
    tagEditionDialog_->setArtists( artistList );
    tagEditionDialog_->setAlbums( albumList );
    #endif

}

//_________________________________________________________
void MainWindow::_processLoadPlaylistsCommand( const Mpd::Command::List& commands )
{

    Debug::Throw( "MainWindow::_processLoadPlaylistsCommand.\n" );

    if( commands.empty() ) return;

    // keep tracks of playlists and tracks
    StoredPlaylistItem::List playlists;
    StoredPlaylistItem::List tracks;

    for( const auto& command:commands )
    {

        if( !command.success() ) continue;

        const auto& commandString( command.commandList().front() );
        if( commandString.startsWith( "listplaylists" ) )
        {
            // get reply
            const auto& reply( command.reply() );

            // time returned by mpd is utc. need to convert to local
            auto calcUtcOffset = []()
            {
                QDateTime now = QDateTime::currentDateTime();
                QDateTime nowUtc = now.toUTC();
                return static_cast<time_t>( 3600*( 24*(now.date().day() - nowUtc.date().day()) + (now.time().hour() - nowUtc.time().hour() ) ) );
            };

            static const time_t utcToLocal = calcUtcOffset();

            QRegExp regExp(  "playlist:\\s+(.*)\\nLast-Modified:\\s+(\\d+)-(\\d+)-(\\d+)T(\\d+):(\\d+):(\\d+)" );
            regExp.setMinimal( true );

            for( int position = 0; ( position = reply.indexOf( regExp, position ) ) >= 0; position += regExp.matchedLength() )
            {
                const auto playlist = regExp.cap(1);
                auto item( StoredPlaylistItem( StoredPlaylistItem::Type::Playlist, playlist ) );

                // stored last modified
                auto timeStamp = TimeStamp()
                    .setYear( regExp.cap(2).toInt() )
                    .setMonth( regExp.cap(3).toInt() )
                    .setDay( regExp.cap(4).toInt() )
                    .setHours( regExp.cap(5).toInt() )
                    .setMinutes( regExp.cap(6).toInt() )
                    .setSeconds( regExp.cap(7).toInt() );

                timeStamp.makeTime();
                timeStamp = TimeStamp( timeStamp.unixTime() + utcToLocal );
                item.setModification( timeStamp );

                playlists.append( item );
            }

        } else if( commandString.startsWith( "listplaylistinfo" ) ) {

            // capture playlist
            QRegExp regExp( "listplaylistinfo \"(.*)\"");
            if( commandString.indexOf( regExp ) < 0 ) continue;
            const auto playlist = regExp.cap(1);

            // read tracks
            auto localTracks( TrackInformation::Helper::list<StoredPlaylistItem>( command.reply() ) );

            // assign playlist type and index
            int index = 0;
            for( auto& track:localTracks )
            {
                track.setType( StoredPlaylistItem::Type::Track );
                track.setPlaylist( playlist );
                track.setTrackIndex( index++ );
            }

            // append to track list
            tracks.append( localTracks );

        }

    }

    // lock playlists widget and save expanded items
    sidePanelWidget_->storedPlaylistWidget().lock();
    sidePanelWidget_->storedPlaylistWidget().saveExpandedItems();

    // assign playlists to model
    if( playlists.isEmpty() )
    {

        sidePanelWidget_->storedPlaylistWidget().model().clear();
        sidePanelWidget_->storedPlaylistWidget().showNotification();

    } else {

        std::sort( playlists.begin(), playlists.end() );
        sidePanelWidget_->storedPlaylistWidget().hideNotification();
        sidePanelWidget_->storedPlaylistWidget().model().set( playlists );

    }

    if( !tracks.empty() )
    {
        std::sort( tracks.begin(), tracks.end() );
        sidePanelWidget_->storedPlaylistWidget().updatePlaylists( tracks );

        // also update covers
        CollectionItem::List collectionItems;
        std::transform( tracks.begin(), tracks.end(), std::back_inserter( collectionItems ),
            []( const StoredPlaylistItem& item ) { return CollectionItem( item ); } );
        std::sort( collectionItems.begin(), collectionItems.end() );
        _updateCovers( collectionItems );

    }

    // restore expanded items and unlock
    sidePanelWidget_->storedPlaylistWidget().restoreExpandedItems();
    sidePanelWidget_->storedPlaylistWidget().unlock();

}

//_________________________________________________________
void MainWindow::_processFindTracksCommand( const Mpd::Command& command )
{

    Debug::Throw( "MainWindow::_processFindTracksCommand.\n" );

    // parse arguments
    const auto& commandString( command.commandList().front() );

    // get artist
    QRegExp artistRegExp( "artist\\s+\"(.*)\"" );
    artistRegExp.setMinimal( true );
    if( commandString.indexOf( artistRegExp ) < 0 ) return;
    const auto artist = artistRegExp.cap(1);

    // get album
    QRegExp albumRegExp( "album\\s+\"(.*)\"" );
    albumRegExp.setMinimal( true );
    const bool hasAlbum( commandString.indexOf( albumRegExp ) >= 0 );
    const auto album = hasAlbum ? albumRegExp.cap(1):QString();

    // get disc
    QRegExp discRegExp( "disc\\s+\"(.*)\"" );
    discRegExp.setMinimal( true );
    const bool hasDisc( commandString.indexOf( discRegExp ) >= 0 );
    const auto disc = hasDisc ? discRegExp.cap(1).toInt():-1;

    // parse tracks from reply
    auto items( TrackInformation::Helper::list<CollectionItem>( command.reply() ) );
    std::sort( items.begin(), items.end() );

    // get albums if not already set
    if( !hasAlbum )
    {
        sidePanelWidget_->collectionWidget().lock();

        // split items by artist and assign albums to model
        sidePanelWidget_->collectionWidget().updateArtists( items );

        // split items by album and assign tracks to model
        sidePanelWidget_->collectionWidget().updateAlbums( items );

        // update covers
        _updateCovers( items );

        sidePanelWidget_->collectionWidget().unlock();

    } else {

        // find parent index
        CollectionItem parentItem( CollectionItem::Type::Album, artist, album );
        if( hasDisc ) parentItem.setDisc( disc );

        if( !sidePanelWidget_->collectionWidget().model().index( parentItem ).isValid() ) return;

        // assign to model
        sidePanelWidget_->collectionWidget().lock();
        sidePanelWidget_->collectionWidget().model().set( parentItem, items );
        sidePanelWidget_->collectionWidget().unlock();

    }

    // check option to see if needs insertion in playlist
    int position( command.optionList().indexOf( "row" ) );
    if( position >= 0 )
    {

        // make sure items are sorted
        std::sort( items.begin(), items.end() );

        // get row from arguments
        int row = command.optionList()[position+1].toInt();

        QStringList commandStringList;
        std::transform( items.begin(), items.end(), std::front_inserter( commandStringList ),
            [&row]( const CollectionItem& item )
            { return QString( "addid \"%1\" %2" ).arg( item.file() ).arg( row ); } );

        emit commandRequested( Mpd::Command( Mpd::Command::Type::AddTracks, commandStringList ) );

    }

    // check deletion command
    if( command.optionList().contains( "delete" ) && deleteTracksDialog_ )
    { deleteTracksDialog_->addItems( items ); }

    #if WITH_TAGLIB
    // check tag edition commands
    if( command.optionList().contains( "addTag" ) )
    { tagEditionDialog_->addFiles( Local::fileList( items ) ); }

    if( command.optionList().contains( "endTagEdition" ) && tagEditionDialog_->readFiles() && tagEditionDialog_->exec() )
    {
        auto path( tagEditionDialog_->location() );
        if( !( path.exists() && path.isDirectory() ) ) path = path.path();

        // music directory
        File musicDirectory( XmlOptions::get().raw( "MUSIC_DIR" ) );
        musicDirectory.addTrailingSlash();

        // request collection update
        if( path.startsWith( musicDirectory ) )
        { updateCollection( path.get().mid( musicDirectory.get().size() ) ); }
    }
    #endif

}

//_________________________________________________________
void MainWindow::_processFindTrackCommand( const Mpd::Command& command )
{

    Debug::Throw( "MainWindow::_processFindTrackCommand.\n" );

    // find tracks
    auto tracks( TrackInformation::Helper::list<TrackInformation>( command.reply() ) );
    if( tracks.size() != 1 ) return;

    // check option to see if needs insertion in playlist
    int position( command.optionList().indexOf( "row" ) );
    if( position >= 0 )
    {
        // get row from arguments
        int row = command.optionList()[position+1].toInt();
        emit commandRequested( Mpd::Command( Mpd::Command::Type::AddTracks, QString( "addid \"%1\" %2" ).arg( tracks.front().file() ).arg( row ) ) );
    }

}

//_________________________________________________________
void MainWindow::_processCollectionSearchCommand( const Mpd::Command& command )
{

    Debug::Throw( "MainWindow::_processCollectionSearchCommand.\n" );

    // do nothing if search is no more active
    if( !searchActive_ ) return;

    // parse arguments
    const auto& commandString( command.commandList().front() );

    // get searched string
    QRegExp regExp( "search\\s+(\\S+)\\s+\"(.+)\"" );
    regExp.setMinimal(true);

    if( commandString.indexOf( regExp ) < 0 ) return;

    const auto tag = regExp.cap(1);
    const auto searchedString = regExp.cap(2);

    // parse tracks from reply
    auto items( TrackInformation::Helper::list<CollectionItem>( command.reply() ) );
    std::sort( items.begin(), items.end() );

    // remove tracks that do not match searched string, for requested tag
    items.erase(
        std::remove_if( items.begin(), items.end(),
        [this, &searchedString]( const CollectionItem& item )
        {
            if( ( searchMask_ & TrackInformation::Artist ) && item.artist().contains( searchedString, Qt::CaseInsensitive ) ) return false;
            else if( ( searchMask_ & TrackInformation::Album ) && item.album().contains( searchedString, Qt::CaseInsensitive ) ) return false;
            else if( ( searchMask_ & TrackInformation::Title ) && item.title().contains( searchedString, Qt::CaseInsensitive ) ) return false;
            else return true;
        }), items.end() );

    // lock collection widget and save expanded items
    sidePanelWidget_->collectionWidget().lock();
    sidePanelWidget_->collectionWidget().saveExpandedItems();

    // get artists
    auto artists( items );
    artists.erase(
        std::unique( artists.begin(), artists.end(), TrackInformation::SameArtistBinaryFTor() ),
        artists.end() );

    for( auto& item:artists )
    { item.setType( CollectionItem::Type::Artist ); }

    // assign to model
    sidePanelWidget_->collectionWidget().model().set( artists );

    // split items by artist, assign albums to model
    sidePanelWidget_->collectionWidget().updateArtists( items );

    // split items by album, assign tracks to model
    sidePanelWidget_->collectionWidget().updateAlbums( items );

    // update covers
    _updateCovers( items );

    // restore expanded items and unlock
    sidePanelWidget_->collectionWidget().restoreExpandedItems();
    sidePanelWidget_->collectionWidget().unlock();

}

//_________________________________________________________
void MainWindow::_doAddFromCollection( const CollectionItem::List& items, int row )
{

    Debug::Throw( "MainWindow::_doAddFromCollection.\n" );

    // check items
    if( items.empty() ) return;

    if( searchActive_ )
    {

        // when search is active, we use only items that are in the model
        CollectionItem::List expandedItems;
        for( const auto& item:items )
        {
            if( item.type() == CollectionItem::Type::Track ) expandedItems.append( item );
            else {

                // copy all track children to expanded indexes
                auto children( sidePanelWidget_->collectionWidget().model().children( item ) );
                std::copy(
                    children.begin(),
                    std::remove_if( children.begin(), children.end(), CollectionItem::DifferentTypeFTor( CollectionItem::Type::Track ) ),
                    std::back_inserter( expandedItems ) );
            }
        }

        // generate command
        QStringList commandStringList;
        std::transform( expandedItems.begin(), expandedItems.end(), std::front_inserter( commandStringList ),
            [&row]( const CollectionItem& item ) { return QString( "addid \"%1\" %2" ).arg( item.file() ).arg( row ); } );

        emit commandRequested( Mpd::Command( Mpd::Command::Type::AddTracks, commandStringList ) );

    } else {

        // store find commands (for adding albums or artists)
        Mpd::Command::List commandList;
        QStringList commandStringList;

        // loop over items
        for( const auto& item:items )
        {
            switch( item.type() )
            {
                case CollectionItem::Type::Track:
                commandStringList.append( QString( "addid \"%1\" %2" ).arg( item.file() ).arg( row++ ) );
                break;

                case CollectionItem::Type::Artist:
                {
                    auto command( Local::findArtistCommand( item ) );
                    command.setOptionList( { "row", QString::number( row ) } );

                    // find commands are  added in reverse order, in order to preserve ordering
                    commandList.prepend( command );
                    break;
                }

                case CollectionItem::Type::Album:
                {

                    auto command( Local::findAlbumCommand( item ) );
                    command.setOptionList( { "row", QString::number( row ) } );

                    // find commands are added in reverse order, in order to preserve ordering
                    commandList.prepend( command );
                    break;
                }

                default: break;
            }

        }

        // add single track addition command
        if( !commandStringList.empty() )
        { commandList.append( Mpd::Command( Mpd::Command::Type::AddTracks, commandStringList ) ); }

        // send add commands
        if( !commandList.empty() )
        { emit commandsRequested( commandList ); }

    }

}

//_________________________________________________________
void MainWindow::_doAddFromStoredPlaylist( const StoredPlaylistItem::List& items, int row )
{

    Debug::Throw( "MainWindow::_doAddFromStoredPlaylist.\n" );

    // check items
    if( items.empty() ) return;

    // store find commands (for adding albums or artists)
    Mpd::Command::List commandList;
    QStringList commandStringList;

    // loop over items
    for( const auto& item:items )
    {
        switch( item.type() )
        {
            case StoredPlaylistItem::Type::Track:
            commandStringList.append( QString( "addid \"%1\" %2" ).arg( item.file() ).arg( row++ ) );
            break;

            case StoredPlaylistItem::Type::Playlist:
            {

                auto command( Local::findPlaylistTracksCommand( item ) );
                command.setOptionList( { "row", QString::number( row ) } );

                // find commands are added in reverse order, in order to preserve ordering
                commandList.prepend( command );
                break;
            }

            default: break;
        }

    }

    // add single track addition command
    if( !commandStringList.empty() )
    { commandList.append( Mpd::Command( Mpd::Command::Type::AddTracks, commandStringList ) ); }

    // send add commands
    if( !commandList.empty() )
    { emit commandsRequested( commandList ); }

}

//_________________________________________________________
#if WITH_CDIO_PARANOIA
void MainWindow::_doAddFromAudioDevice( const TrackInformation::List& tracks, int row )
{

    Debug::Throw( "MainWindow::_doAddFromAudioDevice.\n" );

    // check items
    if( tracks.empty() ) return;

    // generate command
    QStringList commandStringList;
    std::transform( tracks.begin(), tracks.end(), std::front_inserter( commandStringList ),
        [&row]( const TrackInformation& track ) { return QString( "addid \"%1\" %2" ).arg( track.file() ).arg( row ); } );

    emit commandRequested( Mpd::Command( Mpd::Command::Type::AddTracks, commandStringList ) );
}
#endif

//_______________________________________________________________________
void MainWindow::_updateCovers( const CollectionItem::List& items )
{
    // copy list
    auto copy( items );

    // keep one entry per album
    copy.erase( std::unique( copy.begin(), copy.end(), TrackInformation::SameAlbumBinaryFTor() ), copy.end() );

    // assign correct type and update cover
    for( auto item:copy )
    {
        item.setType( CollectionItem::Type::Album );
        if( !collectionItemKeys_.contains( item.hashKey() ) )
        {
            emit requestCover( item );
            collectionItemKeys_.insert( item.hashKey() );
        }
    }

}
