#ifndef MpdCommand_h
#define MpdCommand_h

/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "Counter.h"
#include "CppUtil.h"

#include <QHash>
#include <QList>
#include <QString>
#include <QStringList>
#include <QTextStream>

namespace Mpd
{

    class Command final: private Base::Counter<Command>
    {

        public:

        using List = QList<Command>;

        enum class Type
        {
            Undefined,

            // collection handling
            LoadCollection,
            UpdateCollection,
            SearchCollection,

            // tracks finding
            FindTrack,
            FindTracks,

            // playback control
            Play,
            Resume,
            Stop,
            Pause,
            NextTrack,
            PreviousTrack,
            ChangeVolume,
            Seek,
            ChangeRepeatMode,

            // playlist handling
            PlaylistInfo,
            ClearPlaylist,
            SavePlaylist,
            AddTracks,
            RemoveTracks,
            MoveTracks,

            // stored playlist
            LoadPlaylists,
            StoredPlaylistInfo,

            // mpd status
            Status,
            Statistics,

            // kill
            Kill

        };

        //* constructor
        explicit Command();

        //* constructor
        explicit Command( Type, const QStringList& );

        #if QT_VERSION < 0x040800
        //* constructor
        explicit Command( Type, std::initializer_list<QString>&& );
        #endif

        //* constructor
        explicit Command( Type, const QString& );

        //*@name accessors
        //@{

        //* unique id
        int id() const
        { return id_; }

        //* type
        Type type() const
        { return type_; }

        //* map Type to command name
        static QString name( Type type )
        { return _commandNames()[type]; }

        //* map Type to command name
        QString name() const
        { return name( type_ ); }

        //* commands
        const QStringList& commandList() const
        { return commandList_; }

        //*options
        const QStringList& optionList() const
        { return optionList_; }

        //* reply
        const QString& reply() const
        { return reply_; }

        //* success
        bool success() const
        { return success_; }

        //* for multiple line commands, return subcommands, associated replies, and success
        List subCommands() const;

        //@}

        //*@name modifiers
        //@{

        //* clear options
        void clearOptionList()
        { optionList_.clear(); }

        //* set options
        void setOptionList( const QStringList& value )
        { optionList_ = value; }

        //* add option
        void addOption( const QString& value )
        { optionList_.append( value ); }

        //* reply
        void setReply( const QString& value )
        { reply_ = value; }

        //* success
        void setSuccess( bool value )
        { success_ = value; }

        //@}

        private:

        //* static id counter
        static int& _counter();

        //* map types to command name
        using NameMap = QHash<Type,QString>;

        //* command names
        static const NameMap& _commandNames();

        //* unique id
        int id_ = 0;

        //* type
        Type type_ = Type::Undefined;

        //* list of text commands
        QStringList commandList_;

        //* options
        QStringList optionList_;

        //* reply
        QString reply_;

        //* true if command was succesful
        bool success_ = false;

        //* streamer
        friend QTextStream& operator << (QTextStream&, const Command& );

    };

    using ::qHash;

}

#endif
