/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "MpdIdleInterface.h"

#include "Command.h"
#include "Debug.h"
#include "MpdBaseSocket.h"
#include "MpdServerConfiguration.h"

#include <signal.h>
#include <QDir>
#include <QRegExp>

namespace Mpd
{

    //______________________________________________
    IdleInterface::IdleInterface( QObject* parent ):
        QObject( parent ),
        Counter( "Mpd::IdleInterface" )
    {}

    //______________________________________________
    IdleInterface::~IdleInterface() = default;

    //______________________________________________
    void IdleInterface::connect()
    {
        Debug::Throw( "Mpd::IdleInterface::connect.\n" );

        socket_.reset( BaseSocket::create( this, serverConfiguration_ ) );
        connect( socket_.get(), SIGNAL(disconnected()), SLOT(_socketDisconnected()) );
        connect( socket_.get(), SIGNAL(readyRead()), SLOT(_readFromSocket()) );
        connect( socket_.get(), SIGNAL(error()), SIGNAL(socketError()) );

        socket_->connect( serverConfiguration_ );
    }

    //______________________________________________
    void IdleInterface::disconnect()
    {
        Debug::Throw( "Mpd::IdleInterface::disconnect.\n" );

        if( socket_ )
        {
            socket_->disconnect();
            socket_.reset();
        }

    }

    //______________________________________________
    void IdleInterface::_readFromSocket()
    {

        //* read
        auto array( socket_->read( socket_->bytesAvailable() ) );
        buffer_.append( QString::fromUtf8( array ) );

        //* parse buffer to look for end reply
        QRegExp regexp( "(\n|^)(OK|ACK).*\n" );
        regexp.setMinimal(true);
        auto position( buffer_.indexOf( regexp ) );
        if( position >= 0 )
        {

            // store reply
            const auto reply( buffer_.left( position+regexp.matchedLength() ) );
            Debug::Throw() << "IdleInterface::_readFromSocket - reply: " << reply << endl;
            _processReply( reply );

            // truncate buffer
            buffer_ = buffer_.mid( position+regexp.matchedLength() );

            // re-run idle command
            socket_->write( u8"idle\n" );

        }

    }

    //______________________________________________
    void IdleInterface::_socketDisconnected()
    {
        Debug::Throw( "IdleInterface::_socketDisconnected.\n" );
        emit socketError();
        connect();
    }

    //______________________________________________
    bool IdleInterface::_isSocketConnected() const
    { return socket_ && socket_->isConnected(); }

    //______________________________________________
    void IdleInterface::_processReply( const QString& reply )
    {

        // parse to find the changed subsystems
        QRegExp regExp( "changed:\\s+(\\w+)\n" );
        QStringList subsystems;
        for( int position = 0; ( position = reply.indexOf( regExp, position ) ) >= 0; position += regExp.matchedLength() )
        { subsystems.append( regExp.cap(1) ); }

        Debug::Throw() << "IdleInterface::_processReply - subsystems: " << subsystems.join( " " ) << endl;

        // ask for status
        if( subsystems.contains( QLatin1String( "player" ) ) || subsystems.contains( QLatin1String( "update" ) ) ) emit requestUpdateStatus();
        if( subsystems.contains( QLatin1String( "playlist" ) ) ) emit requestUpdatePlaylist();
        if( subsystems.contains( QLatin1String( "database" ) ) ) emit requestLoadCollection();
        if( subsystems.contains( QLatin1String( "stored_playlist" ) ) ) emit requestLoadStoredPlaylists();
    }

}
