/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "PlaylistDelegate.h"

#include "Color.h"
#include "Debug.h"
#include "IconEngine.h"
#include "IconNames.h"
#include "Singleton.h"
#include "TrackInformationModel.h"
#include "TrackInformation.h"
#include "XmlOptions.h"

#include <QApplication>
#include <QStyleOptionViewItem>
#include <QPainter>

#include <algorithm>
#include <numeric>

namespace Local
{

    //_________________________________________________________
    QString trackTitle( const TrackInformation& track )
    {
        QString out;
        if( track.title().isEmpty() || track.track() <= 0 ) out += track.titleOrFile();
        else out = QString().sprintf( "%02i ", track.track() ) + track.title();
        return out;
    }

    //_________________________________________________________
    QString albumBuffer( const TrackInformation& track )
    {
        // setup text
        const auto album = track.album().isEmpty() ? TrackInformation::UnknownAlbum:track.album();
        const auto artist = track.artist().isEmpty() ? TrackInformation::UnknownArtist:track.artist();
        return QString( "%1\n%2" ).arg( album, artist );
    }

    //_________________________________________________________
    QString durationString( const TrackInformation& track )
    { return track.duration() > 0 ? track.durationString():QString(); }

}

//_________________________________________________________
PlaylistDelegate::PlaylistDelegate( QObject *parent ):
    QAbstractItemDelegate( parent ),
    Counter( "PlaylistDelegate" )
{

    Debug::Throw( "PlaylistDelegate::PlaylistDelegate.\n" );

    // configuration
    connect( Base::Singleton::get().application(), SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    _updateConfiguration();

}

//_________________________________________________________
void PlaylistDelegate::paint( QPainter *painter, const QStyleOptionViewItem &option, const QModelIndex &index ) const
{

    // check index
    if( !index.isValid() ) return;
    if( index.column() != 0 ) return;

    painter->save();

    // selection
    QStyleOptionViewItemV4 viewOption(option);
    viewOption.showDecorationSelected = true;
    auto style = viewOption.widget ? viewOption.widget->style() : QApplication::style();
    style->drawPrimitive(QStyle::PE_PanelItemViewItem, &viewOption, painter, viewOption.widget);

    // modify color
    QPalette::ColorGroup colorGroup = option.state & QStyle::State_Enabled ? QPalette::Normal : QPalette::Disabled;
    if( colorGroup == QPalette::Normal && !(option.state & QStyle::State_Active) ) colorGroup = QPalette::Inactive;

    QColor foreground = ( option.state & QStyle::State_Selected ) ?
        option.palette.color( colorGroup, QPalette::HighlightedText ):
        option.palette.color( colorGroup, QPalette::Text );

    painter->setPen( foreground );

    // layout text
    // get model and track
    const auto model( static_cast<const TrackInformationModel*>( index.model() ) );
    const auto track( model->get( index ) );

    // decide whether one needs to also render the album
    bool needsAlbum = _needsAlbum( *model, index );

    auto textRect( option.rect.adjusted( margin_, margin_, -margin_, -margin_ ) );
    textRect.setLeft( textRect.left() + iconSize_ + spacing_ );
    textRect = style->visualRect( option.direction, option.rect, textRect );

    if( needsAlbum )
    {

        // setup font
        auto boldFont = option.font;
        boldFont.setBold( true );

        // setup albumt text rect
        const auto albumBuffer( Local::albumBuffer( track ) );
        auto albumRect = QFontMetrics( boldFont ).boundingRect( albumBuffer );
        albumRect.setHeight( std::max( albumRect.height(), coverSize_ ) );
        albumRect.moveTopLeft( textRect.topLeft() );

        auto coverRect = albumRect;

        // adjust album rect
        albumRect.setWidth( textRect.width() );
        albumRect.setLeft( albumRect.left() + coverSize_ + spacing_ );

        painter->setFont( boldFont );
        painter->drawText( albumRect, Qt::AlignCenter, albumBuffer );

        // adjust cover rect
        coverRect.setWidth( coverSize_ );
        coverRect.setHeight( coverSize_ );
        coverRect.moveTop( coverRect.top() + (albumRect.height() - coverRect.height())/2 );

        // get icon
        auto icon( model->icon( track ) );
        if( icon.isNull() )
        {

            icon = IconEngine::get( IconNames::Track );
            painter->drawPixmap( coverRect, icon.pixmap( QSize( coverSize_, coverSize_ ) ) );

        } else {

            // drop shadow
            painter->save();
            painter->setPen( QPen( Base::Color( option.palette.color( QPalette::Shadow ) ).addAlpha( 0.3 ).get(), 2 ) );
            painter->setRenderHints(QPainter::Antialiasing );

            auto shadowRect = QRectF( coverRect ).adjusted( 0.5, 0.5, -0.5, -0.5 ).translated( 0.5, 0.5 );
            painter->drawRect( shadowRect );
            painter->restore();

            // cover
            painter->drawPixmap( coverRect, icon.pixmap( QSize( coverSize_, coverSize_ ) ) );

        }

        // update text rect
        textRect.setTop( albumRect.bottom() + 2*spacing_ - 1 );

    }

    // render track title and length
    // setup font
    painter->setFont( option.font );

    // setup text
    const auto trackBuffer = Local::trackTitle( track );
    const auto trackDurationBuffer = Local::durationString( track );

    // render
    auto copy( option );
    copy.rect = textRect;
    _renderTextLeftRight( painter, copy, trackBuffer, trackDurationBuffer );

    // render icon for current track
    if( index == model->currentTrackIndex() )
    {

        // also draw state icon from model
        const auto& icon( model->stateIcon() );
        if( !icon.isNull() )
        {
            // icon rect
            auto iconRect =
                QRect( option.rect.bottomLeft(), QSize( iconSize_, iconSize_ ) )
                .translated( margin_, -iconSize_ - margin_ + 1 );

            iconRect = style->visualRect( option.direction, option.rect, iconRect );

            QIcon::Mode iconMode = QIcon::Normal;
            if( !( option.state & QStyle::State_Enabled ) ) iconMode = QIcon::Disabled;
            else if( ( option.state & QStyle::State_Active ) && ( option.state & QStyle::State_Selected ) ) iconMode = QIcon::Selected;

            // get the relevant pixmap
            painter->drawPixmap( iconRect, icon.pixmap( QSize( iconSize_, iconSize_ ), iconMode ) );

        }

    }

    painter->restore();

}

//_________________________________________________________
QSize PlaylistDelegate::sizeHint( const QStyleOptionViewItem &option, const QModelIndex &index ) const
{

    if( !index.isValid() ) return QSize();
    if( index.column() != 0 ) return QSize();

    // get model and track
    const auto& model( static_cast<const TrackInformationModel*>( index.model() ) );
    const auto& track( model->get( index ) );
    const bool needsAlbum( _needsAlbum( *model, index ) );

    // get needed size for track number, title and duration
    QSize textSize = QFontMetrics( option.font ).boundingRect(  QString( "%1 %2" ).arg( Local::trackTitle( track ), Local::durationString( track ) ) ).size();

    // make sure there is enough vertical space for icon
    textSize.setHeight( std::max( iconSize_, textSize.height() ) );

    // add space for album title if needed
    if( needsAlbum )
    {
        auto boldFont( option.font );
        boldFont.setBold( true );

        // setup text and get bounding size
        const auto albumSize = QFontMetrics( boldFont ).boundingRect(  Local::albumBuffer( track ) ).size();
        textSize.setWidth( std::max( textSize.width(), albumSize.width() + spacing_ + coverSize_ ) );
        textSize.setHeight( textSize.height() + std::max( albumSize.height(), coverSize_ ) + 2*spacing_ );
    }

    // add horizontal space for icon
    textSize.setWidth( textSize.width() + iconSize_ + spacing_ );

    // add margins
    return textSize + QSize( 2*margin_, 2*margin_ );
}

//_________________________________________________________
bool PlaylistDelegate::isInAlbum( const QPoint& position, const QStyleOptionViewItem &option, const QModelIndex& index ) const
{

    const auto model( static_cast<const TrackInformationModel*>( index.model() ) );
    if( !_needsAlbum( *model, index ) ) return false;

    const auto& track( model->get( index ) );
    const auto size( sizeHint( option, index ) );

    QFontMetrics fontMetrics( option.font );
    auto textHeight = fontMetrics.boundingRect( track.track() ).height();

    return position.y() < size.height() - textHeight - margin_ - spacing_;

}

//_________________________________________________________
bool PlaylistDelegate::_needsAlbum( const TrackInformationModel& model, const QModelIndex& index ) const
{
    if( !model.contains( index ) ) return false;
    const auto& track( model.get( index ) );

    QModelIndex previousIndex( model.index( index.row()-1, index.column() ) );
    if( !model.contains( previousIndex ) ) return true;

    const auto& previousTrack( model.get( previousIndex ) );
    return previousTrack.album() != track.album();

}

//_________________________________________________________
void PlaylistDelegate::_renderTextLeftRight( QPainter* painter, const QStyleOptionViewItem &option, const QString& leftText, const QString& rightText ) const
{

    // assign font
    painter->setFont( option.font );

    QRect leftTextRect( option.rect );
    if( !rightText.isEmpty() )
    {

        // draw right text
        const auto rightTextRect = painter->boundingRect( option.rect, Qt::AlignRight|Qt::AlignVCenter, rightText );
        painter->drawText( rightTextRect, Qt::AlignRight|Qt::AlignVCenter, rightText );

        // compute left text rect
        if( option.direction == Qt::LeftToRight ) leftTextRect.setRight( rightTextRect.left() - spacing_ );
        else leftTextRect.setLeft( rightTextRect.right() + spacing_ );

    }

    // draw left text
    painter->drawText( leftTextRect, Qt::AlignLeft|Qt::AlignVCenter, QFontMetrics( option.font ).elidedText( leftText, Qt::ElideRight, leftTextRect.width() ) );
}

//________________________________________________________
void PlaylistDelegate::_updateConfiguration()
{
    Debug::Throw( "PlaylistDelegate::_updateConfiguration.\n" );
    coverSize_ = XmlOptions::get().get<int>( "PLAYLIST_ICON_SIZE" );
}

