/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "PlaylistWriter.h"

#include "CppUtil.h"
#include "Debug.h"
#include "XmlOptions.h"

#include <QDomDocument>
#include <QTextStream>
#include <QUrl>

//_____________________________________________________________________
PlaylistWriter::Format PlaylistWriter::playlistFormat( const File& file )
{
    const auto extension( file.extension().get().toLower() );
    if( extension == QLatin1String( "m3u" ) ) return Format::M3u;
    else if( extension == QLatin1String( "xspf" ) ) return Format::Xspf;
    else return defaultFormat();
}

//_____________________________________________________________________
QString PlaylistWriter::formatExtension( PlaylistWriter::Format format )
{
    switch( format )
    {
        case Format::M3u: return QLatin1String( "m3u" );
        case Format::Xspf: return QLatin1String( "xspf" );
        default: return QString();
    }
}

//_____________________________________________________________________
PlaylistWriter::Format PlaylistWriter::defaultFormat()
{ return static_cast<PlaylistWriter::Format>( XmlOptions::get().get<int>( "DEFAULT_PLAYLIST_FORMAT" ) ); }

//_____________________________________________________________________
PlaylistWriter::PlaylistWriter():
    Counter( "PlaylistWriter" )
{ Debug::Throw( "PlaylistWriter::PlaylistWriter.\n" ); }

//_____________________________________________________________________
template<>
QByteArray PlaylistWriter::create( const QList<TrackInformation>& tracks, PlaylistWriter::Format format ) const
{
    switch( format )
    {
        case Format::M3u: return _createM3u( tracks );
        case Format::Xspf: return _createXspf( tracks );
        default: return QByteArray();
    }
}

//_____________________________________________________________________
QByteArray PlaylistWriter::_createM3u(  const TrackInformation::List& tracks ) const
{

    QString out;
    QTextStream stream( &out );

    // m3u header
    stream << "#EXTM3U" << endl;

    // music path
    File path;
    for( const auto track:tracks )
    {
        auto file( track.file() );
        if( file.isAbsolute() && !file.exists() ) continue;

        if( useAbsolutePath_ && !file.isAbsolute() )
        {
            if( path.isEmpty() ) path = File( XmlOptions::get().raw( "MUSIC_DIR" ) );
            file.addPath( path );
            if( !file.exists() ) continue;

        }

        // write
        stream
            << "#EXTINF:" << track.duration()
            << ", " << (track.artist().isEmpty() ? TrackInformation::UnknownArtist:track.artist());
        if( !track.title().isEmpty() ) stream << " - " << track.title();
        stream << endl;
        stream << file.get() << endl;
    }

    return out.toUtf8();
}


//_____________________________________________________________________
QByteArray PlaylistWriter::_createXspf(  const TrackInformation::List& tracks) const
{
    QDomDocument document;
    document.appendChild( document.createProcessingInstruction("xml", "version=\"1.0\" encoding=\"UTF-8\"") );

    auto top = document.createElement( "playlist" );
    top.setAttribute( "version", QString::number( 1 ) );
    top.setAttribute( "xmlns", "http://xspf.org/ns/0/" );
    document.appendChild( top );

    top = top.appendChild( document.createElement( "trackList" ) ).toElement();

    File path;

    for( const auto& track:tracks )
    {
        auto file( track.file() );
        if( file.isAbsolute() && !file.exists() ) continue;

        if( useAbsolutePath_ && !file.isAbsolute() )
        {
            if( path.isEmpty() ) path = File( XmlOptions::get().raw( "MUSIC_DIR" ) );
            file.addPath( path );
            if( !file.exists() ) continue;

        }

        auto trackNode = top.appendChild( document.createElement( "track" ) ).toElement();

        // file
        QUrl url;
        url.setScheme( "file" );
        url.setPath( file.get() );
        trackNode
            .appendChild( document.createElement( "location" ) )
            .appendChild( document.createTextNode( url.toEncoded() ) );

        // artist
        if( !track.artist().isEmpty() )
        {
            trackNode
                .appendChild( document.createElement( "artist" ) )
                .appendChild( document.createTextNode( track.artist() ) );
        }

        // album
        if( !track.album().isEmpty() )
        {
            trackNode
                .appendChild( document.createElement( "creator" ) )
                .appendChild( document.createTextNode( track.album() ) );
        }

        // title
        if( !track.title().isEmpty() )
        {
            trackNode
                .appendChild( document.createElement( "title" ) )
                .appendChild( document.createTextNode( track.title() ) );
        }

        // track number
        if( track.track() > 0 )
        {
            trackNode
                .appendChild( document.createElement( "trackNum" ) )
                .appendChild( document.createTextNode( QString::number( track.track() ) ) );
        }

        // duration
        if( track.duration() > 0 )
        {
            trackNode
                .appendChild( document.createElement( "duration" ) )
                .appendChild( document.createTextNode( QString::number( track.duration()*1000 ) ) );
        }

    }

    return document.toByteArray();
}
