#ifndef StoredPlaylistItem_h
#define StoredPlaylistItem_h

/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "Counter.h"
#include "Functors.h"
#include "QOrderedSet.h"
#include "TimeStamp.h"
#include "TrackInformation.h"

#include <QList>

class StoredPlaylistItem: public TrackInformation
{

    public:

    using List = QList<StoredPlaylistItem>;
    using Set = QOrderedSet<StoredPlaylistItem>;

    //* used when LogEntry keyword is not defined
    static const QString MimeType;

    //* default constructor
    explicit StoredPlaylistItem()
    {}

    //* item type
    enum class Type
    {
        Invalid,
        Playlist,
        Track
    };

    //* filled constructor
    explicit StoredPlaylistItem( Type type, QString name = QString() ):
        type_( type )
    { setPlaylist( name ); }

    //* constructor from TrackInformation
    explicit StoredPlaylistItem( const TrackInformation& other ):
        TrackInformation( other ),
        type_( Type::Track )
    {}

    //* constructor from TrackInformation
    explicit StoredPlaylistItem( TrackInformation&& other ):
        TrackInformation( std::move(other) ),
        type_( Type::Track )
    {}

    //* constructor from command reply
    explicit StoredPlaylistItem( const QString& value ):
        TrackInformation( value ),
        type_( Type::Track )
    {}

    //*@name accessors
    //@{

    //* item type
    Type type() const
    { return type_; }

    //* parent type
    Type parentType() const
    { return parentType( type() ); }

    //* parent type
    static Type parentType( Type );

    //* playlist name
    const QString& playlist() const
    { return playlist_; }

    //* short playlist name
    static QString displayName( const QString& );

    //* modification time
    static TimeStamp creation( const QString& );

    //* creation time
    const TimeStamp& creation() const
    { return creation_; }

    //* modification time
    const TimeStamp& modification() const
    { return modification_; }

    //* short playlist name
    const QString& displayName() const
    { return displayName_; }

    //* true if is children of other item
    bool isChild( const StoredPlaylistItem& ) const;

    //* track count
    qint32 trackCount() const
    { return type_ == Type::Playlist ? track():0; }

    //@}

    //*@name modifiers
    //@{

    //* item type
    void setType( Type value )
    { type_ = value; }

    //* track count
    void setTrackCount( qint32 value )
    { if( type_ == Type::Playlist ) setTrack( value ); }

    //* name
    void setPlaylist( const QString& );

    //* set creation time
    void setCreation( const TimeStamp& value )
    { creation_ = value; }

    //* set modification time
    void setModification( const TimeStamp& value )
    { modification_ = value; }

    //@}

    using SameTypeFTor = Base::Functor::UnaryEqual<StoredPlaylistItem, StoredPlaylistItem::Type, &StoredPlaylistItem::type>;
    using DifferentTypeFTor = Base::Functor::UnaryDiffer<StoredPlaylistItem, StoredPlaylistItem::Type, &StoredPlaylistItem::type>;
    using SamePlaylistFTor = Base::Functor::UnaryEqual<StoredPlaylistItem, const QString&, &StoredPlaylistItem::playlist>;

    private:

    //* type
    Type type_ = Type::Invalid;

    //* playlist name
    QString playlist_;
    QString displayName_;

    //* timestamps
    TimeStamp creation_;
    TimeStamp modification_;

    //* equal to operator
    friend bool operator == (const StoredPlaylistItem& lhs, const StoredPlaylistItem& rhs );
    friend bool operator < (const StoredPlaylistItem& lhs, const StoredPlaylistItem& rhs );

};

#endif
