/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "SystemTrayIcon.h"

#include "BaseSvgRenderer.h"
#include "CppUtil.h"
#include "IconEngine.h"
#include "IconSize.h"
#include "Singleton.h"
#include "SvgPlasmaInterface.h"
#include "XmlOptions.h"

#include <QApplication>
#include <QDesktopWidget>
#include <QPainter>

//______________________________________________________
SystemTrayIcon::SystemTrayIcon( QObject* parent ):
    QSystemTrayIcon( parent ),
    Counter( "SystemTrayIcon" )
{

    Debug::Throw( "SystemTrayIcon::SystemTrayIcon.\n" );

    // configuration
    auto application( Base::Singleton::get().application() );
    connect( application, SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    _updateConfiguration();

}

//______________________________________________________
SystemTrayIcon::~SystemTrayIcon() = default;

//______________________________________
void SystemTrayIcon::_updateConfiguration()
{

    Debug::Throw( "SystemTrayIcon::_updateConfiguration.\n" );

    if( XmlOptions::get().get<bool>( "USE_MONOCHROME_SYSTEMTRAY_ICON" ) )
    {

        // try get from plasma interface if needed
        Svg::BaseSvgRenderer renderer;

        #if !defined(Q_OS_WIN)
        if( XmlOptions::get().get<bool>( "SVG_USE_PLASMA_INTERFACE" ) && !plasmaInterface_ )
        {
            plasmaInterface_.reset( new Svg::SvgPlasmaInterface( this ) );
            connect( plasmaInterface_.get(), SIGNAL(themeChanged()), SLOT(_reloadIcon()) );

        }
        #endif

    }

    _reloadIcon();

}

//______________________________________________________
void SystemTrayIcon::_reloadIcon()
{
    Debug::Throw( "SystemTrayIcon::_reloadIcon.\n" );

    if( XmlOptions::get().get<bool>( "USE_MONOCHROME_SYSTEMTRAY_ICON" ) )
    {

        // try get from plasma interface if needed
        Svg::BaseSvgRenderer renderer;

        #if !defined(Q_OS_WIN)
        if( XmlOptions::get().get<bool>( "SVG_USE_PLASMA_INTERFACE" ) && plasmaInterface_ )
        {
            plasmaInterface_->loadTheme();
            renderer.createStyleSheet(
                plasmaInterface_->hasThemePalette() ?
                plasmaInterface_->themePalette():
                QPalette() );
        }
        #endif

        // load icon
        const QString inconName( ":/icons/scalable/musicbox-systray.svg" );

        icon_ = QIcon();
        renderer.load( inconName );
        for( auto&& iter = IconSize::map().constBegin(); iter != IconSize::map().constEnd(); ++iter )
        {
            QPixmap pixmap( IconSize::get( iter.key() ) );
            pixmap.fill( Qt::transparent );
            QPainter painter( &pixmap );
            renderer.render( &painter );
            icon_.addPixmap( pixmap );
        }

    } else {

        // store icons
        icon_ = IconEngine::get( "musicbox-systray" );

    }

    setIcon( icon_ );

}

//__________________________________________________________
bool SystemTrayIcon::_needsReloadOnPaletteChange() const
{ return ( XmlOptions::get().get<bool>( "USE_MONOCHROME_SYSTEMTRAY_ICON" ) && !(plasmaInterface_ && plasmaInterface_->hasThemePalette()) ); }
