
/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "TagLibInterface.h"

#include "Debug.h"

#include <QFile>

#if WITH_TAGLIB
#include <taglib/tpropertymap.h>

static const TagLib::String discTag( "DISCNUMBER" );
#endif



//___________________________________________________
TagLibInterface::TagLibInterface( const File& file ):
    Counter( "TagLibInterface" ),
    file_( file )
    #if WITH_TAGLIB
    , fileRef_( QFile::encodeName(file_.get()).constData() )
    #endif
{ Debug::Throw( "TagLibInterface::TagLibInterface.\n" ); }

//___________________________________________________
bool TagLibInterface::isValid() const
{
    #if WITH_TAGLIB
    return !fileRef_.isNull() && fileRef_.tag();
    #else
    return false;
    #endif
}

//___________________________________________________
TrackInformation TagLibInterface::trackInformation( TrackInformation::Mask mask ) const
{
    Debug::Throw( "TagLibInterface::trackInformation.\n" );

    TrackInformation out;
    out.setFile( file_ );

    if( !isValid() ) return out;

    #if WITH_TAGLIB
    if( mask&TrackInformation::Artist ) out.setArtist( QString::fromUtf8( fileRef_.tag()->artist().toCString( true ) ) );
    if( mask&TrackInformation::Album ) out.setAlbum( QString::fromUtf8( fileRef_.tag()->album().toCString( true ) ) );
    if( mask&TrackInformation::Title ) out.setTitle( QString::fromUtf8( fileRef_.tag()->title().toCString( true ) ) );
    if( mask&TrackInformation::Year ) out.setYear( fileRef_.tag()->year() );
    if( mask&TrackInformation::Track ) out.setTrack( fileRef_.tag()->track() );

    // try read extra tags from property map
    if( mask&TrackInformation::Disc )
    {
        const auto propertyMap( fileRef_.file()->properties() );
        const auto iter( propertyMap.find( discTag ) );
        if( iter != propertyMap.end() && !iter->second.isEmpty() )
        { out.setDisc( iter->second.front().toInt() ); }
    }

    #endif

    return out;
}

//___________________________________________________
void TagLibInterface::setFile( const File& file )
{
    Debug::Throw( "TagLibInterface::setFile.\n" );
    file_ = file;

    #if WITH_TAGLIB
    fileRef_ = TagLib::FileRef( QFile::encodeName(file_.get()).constData() );
    #endif
}

//___________________________________________________
bool TagLibInterface::setTrackInformation( const TrackInformation& trackInformation, TrackInformation::Mask mask )
{

    if( !isValid() ) return false;
    Debug::Throw( "TagLibInterface::setTrackInformation.\n" );

    #if WITH_TAGLIB
    if( !mask ) return true;

    if( mask&TrackInformation::Artist && !trackInformation.artist().isEmpty() ) fileRef_.tag()->setArtist( TagLib::String( trackInformation.artist().toUtf8().constData(), TagLib::String::UTF8 ) );
    if( mask&TrackInformation::Album && !trackInformation.album().isEmpty() ) fileRef_.tag()->setAlbum( TagLib::String( trackInformation.album().toUtf8().constData(), TagLib::String::UTF8 ) );
    if( mask&TrackInformation::Title && !trackInformation.title().isEmpty() ) fileRef_.tag()->setTitle( TagLib::String( trackInformation.title().toUtf8().constData(), TagLib::String::UTF8 ) );
    if( mask&TrackInformation::Year && trackInformation.year() > 0 ) fileRef_.tag()->setYear( trackInformation.year() );
    if( mask&TrackInformation::Track && trackInformation.track() > 0 ) fileRef_.tag()->setTrack( trackInformation.track() );

    // assign extra properties
    if( mask&TrackInformation::Disc && trackInformation.disc() > 0 )
    {
        auto propertyMap( fileRef_.file()->properties() );
        propertyMap.replace( discTag, TagLib::String( QString::number( trackInformation.disc() ).toUtf8().constData(), TagLib::String::UTF8 ) );
        fileRef_.file()->setProperties( propertyMap );
    }

    fileRef_.save();

    return true;
    #else
    return false;
    #endif

}
